# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Launcher script for afl++ based fuzzers."""

# pylint: disable=g-statement-before-imports
try:
  # ClusterFuzz dependencies.
  from clusterfuzz._internal.base import modules
  modules.fix_module_search_paths()
except ImportError:
  pass

import atexit
import collections
import os
import re
import shutil
import signal
import stat
import subprocess
import sys

import six

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.bot.fuzzers import dictionary_manager
from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.bot.fuzzers import options
from clusterfuzz._internal.bot.fuzzers import strategy_selection
from clusterfuzz._internal.bot.fuzzers import utils as fuzzer_utils
from clusterfuzz._internal.bot.fuzzers.afl import constants
from clusterfuzz._internal.bot.fuzzers.afl import stats
from clusterfuzz._internal.bot.fuzzers.afl.fuzzer import write_dummy_file
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.fuzzing import strategy
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.metrics import profiler
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import minijail
from clusterfuzz._internal.system import new_process
from clusterfuzz._internal.system import shell

# Allow 30 minutes to merge the testcases back into the corpus. This matches
# libFuzzer's merge timeout.
DEFAULT_MERGE_TIMEOUT = 30 * 60

BOT_NAME = environment.get_value('BOT_NAME', '')

STDERR_FILENAME = 'stderr.out'

MAX_OUTPUT_LEN = 1 * 1024 * 1024  # 1 MB

# .options file option for the number of persistent executions.
PERSISTENT_EXECUTIONS_OPTION = 'n'

# Grace period for the launcher to complete any processing before it's killed.
# This will no longer be needed when we migrate to the engine interface.
POSTPROCESSING_TIMEOUT = 30


class AflOptionType(object):
  ARG = 0
  ENV_VAR = 1


# Afl options have names and can either be commandline arguments or environment
# variables.
AflOption = collections.namedtuple('AflOption', ['name', 'type'])


class AflConfig(object):
  """Helper class that determines the arguments that should be passed to
  afl-fuzz, environment variables that should be set before running afl-fuzz,
  and the number of persistent executions that should be passed to the target
  Determines these mainly by parsing the .options file for the target."""

  # Mapping of libfuzzer option names to AflOption objects.
  LIBFUZZER_TO_AFL_OPTIONS = {
      'dict':
          AflOption(constants.DICT_FLAG, AflOptionType.ARG),
      'close_fd_mask':
          AflOption(constants.CLOSE_FD_MASK_ENV_VAR, AflOptionType.ENV_VAR),
  }

  def __init__(self):
    """Sets the configs to sane defaults. Use from_target_path if you want to
    use the .options file and possibly .dict to set the configs."""
    self.additional_afl_arguments = []
    self.additional_env_vars = {}
    self.num_persistent_executions = constants.MAX_PERSISTENT_EXECUTIONS

  @classmethod
  def from_target_path(cls, target_path):
    """Instantiates and returns an AFLConfig object. The object is configured
    based on |target_path|."""
    config = cls()
    config.parse_options(target_path)
    config.dict_path = fuzzer_utils.extract_argument(
        config.additional_afl_arguments, constants.DICT_FLAG, remove=False)

    config.use_default_dict(target_path)
    dictionary_manager.correct_if_needed(config.dict_path)

    return config

  def parse_options(self, target_path):
    """Parses a target's .options file (determined using |target_path|) if it
    exists and sets configs based on it."""
    fuzzer_options = options.get_fuzz_target_options(target_path)
    if not fuzzer_options:
      return

    self.additional_env_vars = fuzzer_options.get_env()

    # Try to convert libFuzzer arguments to AFL arguments or env vars.
    libfuzzer_options = fuzzer_options.get_engine_arguments('libfuzzer')
    for libfuzzer_name, value in six.iteritems(libfuzzer_options.dict()):
      if libfuzzer_name not in self.LIBFUZZER_TO_AFL_OPTIONS:
        continue

      afl_option = self.LIBFUZZER_TO_AFL_OPTIONS[libfuzzer_name]
      if afl_option.type == AflOptionType.ARG:
        self.additional_afl_arguments.append('%s%s' % (afl_option.name, value))
      else:
        assert afl_option.type == AflOptionType.ENV_VAR
        self.additional_env_vars[afl_option.name] = value

    # Get configs set specifically for AFL.
    afl_options = fuzzer_options.get_engine_arguments('AFL')
    self.num_persistent_executions = afl_options.get(
        PERSISTENT_EXECUTIONS_OPTION, constants.MAX_PERSISTENT_EXECUTIONS)

  def use_default_dict(self, target_path):
    """Set the dictionary argument in |self.additional_afl_arguments| to
    %target_binary_name%.dict if no dictionary argument is already specified.
    Also update |self.dict_path|."""
    if self.dict_path:
      return

    default_dict_path = dictionary_manager.get_default_dictionary_path(
        target_path)
    if not os.path.exists(default_dict_path):
      return

    self.dict_path = default_dict_path
    self.additional_afl_arguments.append(constants.DICT_FLAG + self.dict_path)


class AflFuzzOutputDirectory(object):
  """Helper class used by AflRunner to deal with AFL's output directory and its
  contents (ie: the -o argument to afl-fuzz)."""

  # AFL usually copies over old units from the corpus to the queue and adds the
  # string 'orig' to the new filename. Therefore we know that testcases
  # containing 'orig' are copied.
  COPIED_FILE_STRING = 'orig'

  TESTCASE_REGEX = re.compile(r'id:\d{6},.+')

  def __init__(self):
    self.output_directory = os.path.join(fuzzer_utils.get_temp_dir(),
                                         'afl_output_dir')

    engine_common.recreate_directory(self.output_directory)

  @classmethod
  def is_testcase(cls, path):
    """Is the path an AFL testcase file or something else."""
    return (os.path.isfile(path) and
            bool(re.match(cls.TESTCASE_REGEX, os.path.basename(path))))

  @property
  def instance_directory(self):
    """Returns afl-fuzz's instance directory."""
    return os.path.join(self.output_directory, constants.DEFAULT_INSTANCE_ID)

  @property
  def queue(self):
    """Returns afl-fuzz's queue directory."""
    return os.path.join(self.instance_directory, 'queue')

  def is_new_testcase(self, path):
    """Determine if |path| is a new unit."""
    # Clearly non-testcases can't be new testcases.
    return (self.is_testcase(path) and
            self.COPIED_FILE_STRING not in os.path.basename(path))

  def count_new_units(self, corpus_path):
    """Count the number of new units (testcases) in |corpus_path|."""
    corpus_files = os.listdir(corpus_path)
    num_new_units = 0

    for testcase in corpus_files:
      if self.is_new_testcase(os.path.join(corpus_path, testcase)):
        num_new_units += 1

    return num_new_units

  def copy_crash_if_needed(self, testcase_path):
    """Copy the first crash found by AFL to |testcase_path| (the input file
    created by run.py).
    """
    crash_paths = list_full_file_paths(
        os.path.join(self.instance_directory, 'crashes'))

    for crash_path in crash_paths:
      # AFL puts a README.txt file in the crashes directory. Just ignore it.
      if self.is_testcase(crash_path):
        shutil.copyfile(crash_path, testcase_path)
        break

  def remove_hang_in_queue(self, hang_filename):
    """Removes the hanging testcase from queue."""
    # AFL copies all inputs to the queue and renames them in the format
    # "id:NUMBER,orig:$hang_filename". So remove that file from the queue.
    # TODO(metzman): What about for the copied inputs without 'orig' in their
    # name that we have been seeing? Does this work anyway because AFL uses the
    # the full name of the file in the queue?
    queue_paths = list_full_file_paths(self.queue)
    hang_queue_path = [
        path for path in queue_paths if path.endswith(hang_filename)
    ][0]

    remove_path(hang_queue_path)

  @property
  def stats_path(self):
    """Returns the path of AFL's stats file: "fuzzer_stats"."""
    return os.path.join(self.instance_directory, 'fuzzer_stats')


class FuzzingStrategies(object):
  """Helper class used by AflRunner classes to decide what strategy to use
  and to record the decision for StatsGetter to use later."""

  # Probability for the level of CMPLOG to set. (-l X)
  CMPLOG_LEVEL_PROBS = [
      ('1', 0.6),  # Level 1
      ('2', 0.3),  # Level 2
      ('3', 0.1),  # Level 3
  ]

  # Probabilty for arithmetic CMPLOG calculations.
  CMPLOG_ARITH_PROB = 0.1

  # Probability for transforming CMPLOG solving.
  CMPLOG_TRANS_PROB = 0.1

  # Probability to disable trimming. (AFL_DISABLE_TRIM=1)
  DISABLE_TRIM_PROB = 0.75

  # Probability for increased havoc intensity. (AFL_EXPAND_HAVOC_NOW=1)
  EXPAND_HAVOC_PROB = 0.5

  # Probability to use the MOpt mutator. (-L0)
  MOPT_PROB = 0.4

  # Probability to use the original afl queue walking mechanism. (-Z)
  QUEUE_OLD_STRATEGY_PROB = 0.2

  # Probability to enable the schedule cycler. (AFL_CYCLE_SCHEDULES=1)
  SCHEDULER_CYCLE_PROB = 0.1

  # Propability which scheduler to select. (-p SCHEDULER)
  SCHEDULER_PROBS = [
      ('fast', .3),
      ('explore', .3),
      ('exploit', .2),
      ('coe', .1),
      ('rare', .1),
  ]

  # TODO(mbarbella): The codepath involving |strategy_dict| and the
  # to_strategy_dict function should be removed when everything is fully
  # converted to the engine pipeline. For now this allows the code to be shared
  # between both cases while still adhering to the new pipeline's API.
  def __init__(self, target_path, strategy_dict=None):
    self.corpus_subset_size = None
    self.candidate_generator = engine_common.Generator.NONE

    # If we have already generated a strategy dict, use that in favor of
    # creating a new pool and picking randomly.
    if strategy_dict is not None:
      self.use_corpus_subset = 'corpus_subset' in strategy_dict
      if self.use_corpus_subset:
        self.corpus_subset_size = strategy_dict['corpus_subset']

      if strategy_dict.get(strategy.CORPUS_MUTATION_RADAMSA_STRATEGY.name) == 1:
        self.candidate_generator = engine_common.Generator.RADAMSA
      elif strategy_dict.get(
          strategy.CORPUS_MUTATION_ML_RNN_STRATEGY.name) == 1:
        self.candidate_generator = engine_common.Generator.ML_RNN
    else:
      strategy_pool = strategy_selection.generate_weighted_strategy_pool(
          strategy_list=strategy.AFL_STRATEGY_LIST,
          use_generator=True,
          engine_name='afl')

      # Select a generator to attempt to use for existing testcase mutations.
      self.candidate_generator = engine_common.select_generator(
          strategy_pool, target_path)

      self.use_corpus_subset = strategy_pool.do_strategy(
          strategy.CORPUS_SUBSET_STRATEGY)

      if self.use_corpus_subset:
        self.corpus_subset_size = engine_common.random_choice(
            engine_common.CORPUS_SUBSET_NUM_TESTCASES)

    self.is_mutations_run = (
        self.candidate_generator != engine_common.Generator.NONE)

    # Generator that is actually used. Initialize to none, change if new
    # testcase mutations are properly generated by the candidate generator.
    self.generator_strategy = engine_common.Generator.NONE

  def to_strategy_dict(self):
    """Convert to a strategy dict in the format used by the engine pipeline."""
    # The decision on whether or not fast cal should be used is made during
    # fuzzing. This function is expected to be called whe preparing for fuzzing.
    strategies_dict = {}

    if self.generator_strategy == engine_common.Generator.RADAMSA:
      strategies_dict[strategy.CORPUS_MUTATION_RADAMSA_STRATEGY.name] = 1
    elif self.generator_strategy == engine_common.Generator.ML_RNN:
      strategies_dict[strategy.CORPUS_MUTATION_ML_RNN_STRATEGY.name] = 1

    if self.use_corpus_subset:
      strategies_dict['corpus_subset'] = self.corpus_subset_size

    return strategies_dict


class AflFuzzInputDirectory(object):
  """Helper class used by AflRunner to deal with the input directory passed to
  afl-fuzz as the -i argument.
  """

  # If the number of input files is less than this, don't bother skipping
  # deterministic steps since it won't take long.
  MIN_INPUTS_FOR_SKIP = 10

  MAX_COPIED_CORPUS_SIZE = 2**30  # 1 GB

  def __init__(self, input_directory, target_path, fuzzing_strategies):
    """Inits AflFuzzInputDirectory.

    Args:
      input_directory: Directory passed to afl-fuzz containing corpus.
      target_path: Path to the fuzzer executable. Used to find seed corpus.
      fuzzing_strategies: fuzzing strategies to use.
    """

    self.input_directory = input_directory
    self.strategies = fuzzing_strategies

    # We only need to use this when a temporary input directory is made.
    # (ie: when there is an oversized testcase in the input).
    self.original_input_directory = None

    engine_common.unpack_seed_corpus_if_needed(
        target_path, self.input_directory, max_bytes=constants.MAX_FILE_BYTES)

    # Ensure there is a usable testcase in the input directory. This is needed
    # because locally (and possibly for new fuzzers on CF) the dummy file is not
    # always in the input directory, which prevents AFL from running.
    if not list_full_file_paths(self.input_directory):
      write_dummy_file(self.input_directory)

  def restore_if_needed(self):
    """Restore the original input directory if self.original_input_directory is
    set. Used to by merge() to get rid of the temporary input directory if it
    exists and merge new units into the original input directory.
    """
    if self.original_input_directory is None:
      return

    # Remove the current input directory as it was only temporary.
    remove_path(self.input_directory)
    self.input_directory = self.original_input_directory
    self.original_input_directory = None


class AflRunnerCommon(object):
  """Afl runner common routines."""

  # Window of time for afl to exit gracefully before we kill it.
  AFL_CLEAN_EXIT_TIME = 10.0

  # Time to wait for SIGTERM handler.
  SIGTERM_WAIT_TIME = 10.0

  # Maximum number of times we will retry fuzzing after fixing an issue.
  MAX_FUZZ_RETRIES = 40

  # Maximum number of times we will retry fuzzing with a strict autocalibrated
  # timeout. After this number of fuzzing retries, if we see a hang we will set
  # the timeout to to '-t' + str(self.MANUAL_TIMEOUT_MILLISECONDS) + '+' to tell
  # AFL to skip testcases that take longer.
  MAX_FUZZ_RETRIES_WITH_STRICT_TIMEOUT = 20

  # The number of times we will retry fuzzing with the deferred fork server
  # after the first testcase hangs. Afterwards we will use
  # AFL_DRIVER_DONT_DEFER, since this is a common symptom of fuzzers that
  # cant use the deferred forkserver.
  MAX_FIRST_HANGS_WITH_DEFERRED_FORKSERVER = 5

  # The timeout we will use if autocalibrating results in too many hangs. This
  # is the maximum autocalibrated timeout afl-fuzz can set.
  MANUAL_TIMEOUT_MILLISECONDS = 5000

  # Regexes used to determine which file caused AFL to quit.
  CRASH_REGEX = re.compile(
      r'Test case \'id\:\d+,orig:(?P<orig_testcase_filename>.*)\' results in a'
      ' crash')

  HANG_REGEX = re.compile(
      r'Test case \'(?P<testcase_filename>.*)\' results in a (hang|timeout)')

  CPU_BIND_ERROR_REGEX = re.compile('PROGRAM ABORT :.*No more free CPU cores')

  # Log messages we format and log as error when afl-fuzz stops running.
  CRASH_LOG_MESSAGE = 'Testcase {0} in corpus causes a crash'
  HANG_LOG_MESSAGE = 'Testcase {0} in corpus causes a hang, retrying without it'

  SHOWMAP_FILENAME = 'afl_showmap_output'
  SHOWMAP_REGEX = re.compile(br'(?P<guard>\d{6}):(?P<hit_count>\d+)\n')

  def __init__(self,
               target_path,
               config,
               testcase_file_path,
               input_directory,
               timeout=None,
               afl_tools_path=None,
               strategy_dict=None):
    """Inits the AflRunner.

    Args:
      target_path: Path to the fuzz target.
      config: AflConfig object.
      testcase_file_path: File to write crashes to.
      input_directory: Corpus directory passed to afl-fuzz.
      afl_tools_path: Path that is used to locate afl-* tools.
    """

    self.target_path = target_path
    self.config = config
    self.testcase_file_path = testcase_file_path
    self._input_directory = input_directory
    self.timeout = timeout

    if afl_tools_path is None:
      afl_tools_path = os.path.dirname(target_path)

    # Set paths to afl tools.
    self.afl_fuzz_path = os.path.join(afl_tools_path, 'afl-fuzz')
    self.afl_showmap_path = os.path.join(afl_tools_path, 'afl-showmap')

    self._afl_input = None
    self._afl_output = None

    self.strategies = FuzzingStrategies(
        target_path, strategy_dict=strategy_dict)

    # Set this to None so we can tell if it has never been set or if it's just
    # empty.
    self._fuzzer_stderr = None

    self.initial_max_total_time = 0

    for env_var, value in six.iteritems(config.additional_env_vars):
      environment.set_value(env_var, value)

    self._showmap_output_path = None
    self.merge_timeout = engine_common.get_merge_timeout(DEFAULT_MERGE_TIMEOUT)
    self.showmap_no_output_logged = False
    self._fuzz_args = []

  @property
  def showmap_output_path(self):
    """Returns the showmap output path."""
    # Initialize _showmap_output_path lazily since MiniJailRunner needs to
    # execute its __init__ before it can be set.
    if self._showmap_output_path is None:
      if environment.get_value('USE_MINIJAIL'):
        self._showmap_output_path = os.path.join(self.chroot.directory,
                                                 self.SHOWMAP_FILENAME)
      else:
        self._showmap_output_path = os.path.join(fuzzer_utils.get_temp_dir(),
                                                 self.SHOWMAP_FILENAME)

    return self._showmap_output_path

  @property
  def stderr_file_path(self):
    """Returns the file for afl to output stack traces."""
    return os.path.join(fuzzer_utils.get_temp_dir(), STDERR_FILENAME)

  @property
  def fuzzer_stderr(self):
    """Returns the stderr of the fuzzer. Reads it first if it wasn't already
    read. Because ClusterFuzz terminates this process after seeing a stacktrace
    printed, make sure that printing this property is the last code a program
    expects to execute.
    """
    if self._fuzzer_stderr is not None:
      return self._fuzzer_stderr

    try:
      with open(self.stderr_file_path, 'rb') as file_handle:
        stderr_data = utils.decode_to_unicode(
            utils.read_from_handle_truncated(file_handle, MAX_OUTPUT_LEN))

      self._fuzzer_stderr = get_first_stacktrace(stderr_data)
    except IOError:
      self._fuzzer_stderr = ''
    return self._fuzzer_stderr

  def run_single_testcase(self, testcase_path):
    """Runs a single testcase.

    Args:
      testcase_path: Path to testcase to be run.

    Returns:
      A new_process.ProcessResult.
    """
    self._executable_path = self.target_path

    assert not testcase_path.isdigit(), ('We don\'t want to specify number of'
                                         ' executions by accident.')

    self.afl_setup()
    result = self.run_and_wait(additional_args=[testcase_path])
    print('Running command:', engine_common.get_command_quoted(result.command))
    if result.return_code not in [0, 1, -6]:
      logs.log_error(
          'AFL target exited with abnormal exit code: %s.' % result.return_code,
          output=result.output)

    return result

  def set_environment_variables(self):
    """Sets environment variables needed by afl."""
    # Tell afl_driver to duplicate stderr to STDERR_FILENAME.
    # Environment variable names and values that must be set before running afl.
    environment.set_value(constants.FORKSRV_INIT_TMOUT_ENV_VAR,
                          constants.FORKSERVER_TIMEOUT)
    environment.set_value(constants.FAST_CAL_ENV_VAR, 1)
    environment.set_value(constants.IGNORE_UNKNOWN_ENVS_ENV_VAR, 1)
    environment.set_value(constants.SKIP_CRASHES_ENV_VAR, 1)
    environment.set_value(constants.SKIP_CPUFREQ_ENV_VAR, 1)
    environment.set_value(constants.BENCH_UNTIL_CRASH_ENV_VAR, 1)
    environment.set_value(constants.EXPAND_HAVOC_NOW_VAR, 1)
    environment.set_value(constants.STDERR_FILENAME_ENV_VAR,
                          self.stderr_file_path)

  def afl_setup(self):
    """Make sure we can run afl. Delete any files that afl_driver needs to
    create and set any environmnet variables it needs.
    """
    self.set_environment_variables()
    remove_path(self.stderr_file_path)

  @staticmethod
  def set_resume(afl_args):
    """Changes afl_args so afl-fuzz will resume fuzzing rather than restarting.
    """
    return AflRunner.set_input_arg(afl_args, constants.RESUME_INPUT)

  @staticmethod
  def get_arg_index(afl_args, flag):
    for idx, arg in enumerate(afl_args):
      if arg.startswith(flag):
        return idx

    return -1

  @classmethod
  def set_arg(cls, afl_args, flag, value):
    """Sets the afl |flag| to |value| in |afl_args|. If |flag| is already
    in |afl_args|, then the old value is replaced by |value|, otherwise |flag|
    and |value| are added.
    """
    idx = cls.get_arg_index(afl_args, flag)
    if value:
      new_arg = flag + str(value)
    else:
      new_arg = flag

    # Arg is not already in afl_args, add it.
    if idx == -1:
      afl_args.insert(0, new_arg)
    else:
      afl_args[idx] = new_arg

    return afl_args

  @classmethod
  def remove_arg(cls, afl_args, flag):
    idx = cls.get_arg_index(afl_args, flag)
    if idx == -1:
      return
    del afl_args[idx]

  @classmethod
  def set_input_arg(cls, afl_args, new_input_value):
    """Changes the input argument (-i) in |afl_args| to |new_input_value|."""
    return cls.set_arg(afl_args, constants.INPUT_FLAG, new_input_value)

  @classmethod
  def set_timeout_arg(cls, afl_args, timeout_value, skip_hangs=False):
    timeout_value = str(int(timeout_value))
    if skip_hangs:
      timeout_value += '+'

    cls.set_arg(afl_args, constants.TIMEOUT_FLAG, timeout_value)
    return afl_args

  def do_offline_mutations(self):
    """Mutate the corpus offline using Radamsa or ML RNN if specified."""
    if not self.strategies.is_mutations_run:
      return

    target_name = os.path.basename(self.target_path)
    project_qualified_target_name = (
        data_types.fuzz_target_project_qualified_name(utils.current_project(),
                                                      target_name))
    # Generate new testcase mutations according to candidate generator. If
    # testcase mutations are properly generated, set generator strategy
    # accordingly.
    generator_used = engine_common.generate_new_testcase_mutations(
        self.afl_input.input_directory, self.afl_input.input_directory,
        project_qualified_target_name, self.strategies.candidate_generator)

    if generator_used:
      self.strategies.generator_strategy = self.strategies.candidate_generator

    # Delete large testcases created by generators.
    for input_path in shell.get_files_list(self.afl_input.input_directory):
      if os.path.getsize(input_path) >= constants.MAX_FILE_BYTES:
        remove_path(input_path)

  def generate_afl_args(self,
                        afl_input=None,
                        afl_output=None,
                        mem_limit=constants.MAX_MEMORY_LIMIT):
    """Generate arguments to pass to Process.run_and_wait.

    Args:
      afl_input: Initial corpus directory passed as -i parameter to the afl
      tool. Defaults to self.afl_input.input_directory.

      afl_output: Output directory where afl stores corpus and stats, passed as
      -o parameter to the afl tool. Defaults to
      self.afl_output.output_directory.

      mem_limit: Virtual memory limit afl enforces on target binary, passed as
      -m parameter to the afl tool. Defaults to constants.MAX_MEMORY_LIMIT.

    Returns:
      A list built from the function's arguments that can be passed as the
      additional_args argument to Process.run_and_wait.

    """

    if afl_input is None:
      afl_input = self.afl_input.input_directory

    if afl_output is None:
      afl_output = self.afl_output.output_directory

    afl_args = [
        constants.INSTANCE_ID_FLAG + constants.DEFAULT_INSTANCE_ID,
        constants.INPUT_FLAG + afl_input, constants.OUTPUT_FLAG + afl_output,
        constants.MEMORY_LIMIT_FLAG + str(mem_limit)
    ]

    afl_args.extend(self.config.additional_afl_arguments)

    afl_args.extend(
        [self.target_path,
         str(self.config.num_persistent_executions)])

    return afl_args

  def should_try_fuzzing(self, max_total_time, num_retries):
    """Returns True if we should try fuzzing, based on the number of times we've
    already tried, |num_retries|, and the amount of time we have left
    (calculated using |max_total_time|).
    """
    if max_total_time <= 0:
      logs.log_error('Tried fuzzing for {0} seconds. Not retrying'.format(
          self.initial_max_total_time))

      return False

    if num_retries > self.MAX_FUZZ_RETRIES:
      logs.log_error(
          'Tried to retry fuzzing {0} times. Fuzzer is likely broken'.format(
              num_retries))

      return False

    return True

  def run_afl_fuzz(self, fuzz_args):
    """Run afl-fuzz and if there is an input that causes afl-fuzz to hang
    or if it can't bind to a cpu, try fixing the issue and running afl-fuzz
    again. If there is a crash in the starting corpus then report it.
    Args:
      fuzz_args: The arguments passed to afl-fuzz. List may be modified if
        afl-fuzz runs into an error.
    Returns:
      A new_process.ProcessResult.
    """
    # Define here to capture in closures.
    max_total_time = self.initial_max_total_time
    fuzz_result = None

    def get_time_spent_fuzzing():
      """Gets the amount of time spent running afl-fuzz so far."""
      return self.initial_max_total_time - max_total_time

    def check_error_and_log(error_regex, log_message_format):
      """See if error_regex can match in fuzz_result.output. If it can, then it
      uses the match to format and print log_message and return the match.
      Otherwise returns None.
      """
      matches = re.search(error_regex, fuzz_result.output)
      if matches:
        erroring_filename = matches.groups()[0]
        message_format = (
            'Seconds spent fuzzing: {seconds}, ' + log_message_format)

        logs.log(
            message_format.format(
                erroring_filename, seconds=get_time_spent_fuzzing()))

        return erroring_filename
      return None  # else

    num_first_testcase_hangs = 0
    num_retries = 0
    while self.should_try_fuzzing(max_total_time, num_retries):
      # Increment this now so that we can just "continue" without incrementing.
      num_retries += 1

      self.afl_setup()

      # If the target was compiled for CMPLOG we need to set this.
      build_directory = environment.get_value('BUILD_DIR')
      cmplog_build_file = os.path.join(build_directory, 'afl_cmplog.txt')
      if os.path.exists(cmplog_build_file):
        self.set_arg(fuzz_args, constants.CMPLOG_FLAG, self.target_path)

      # If a compile time dictionary was created - load it.
      aflpp_dict_file = os.path.join(build_directory, 'afl++.dict')
      if os.path.exists(aflpp_dict_file):
        self.set_arg(fuzz_args, constants.DICT_FLAG, aflpp_dict_file)

      # In the following section we randomly select different strategies.

      # Randomly select a scheduler.
      self.set_arg(fuzz_args, constants.SCHEDULER_FLAG,
                   rand_schedule(self.strategies.SCHEDULER_PROBS))

      # Randomly set trimming vs no trimming.
      if engine_common.decide_with_probability(
          self.strategies.DISABLE_TRIM_PROB):
        environment.set_value(constants.DISABLE_TRIM_ENV_VAR, 1)

      # Randomly enable expanded havoc mutation.
      if engine_common.decide_with_probability(
          self.strategies.EXPAND_HAVOC_PROB):
        environment.set_value(constants.EXPAND_HAVOC_ENV_VAR, 1)

      # Always CMPLOG only new finds.
      environment.set_value(constants.CMPLOG_ONLY_NEW_ENV_VAR, 1)

      # Randomly set new vs. old queue selection mechanism.
      if engine_common.decide_with_probability(
          self.strategies.QUEUE_OLD_STRATEGY_PROB):
        self.set_arg(fuzz_args, constants.QUEUE_OLD_STRATEGY_FLAG, None)

      # Randomly select the MOpt mutator.
      if engine_common.decide_with_probability(self.strategies.MOPT_PROB):
        self.set_arg(fuzz_args, constants.MOPT_FLAG, '0')

      # Select the CMPLOG level (even if no cmplog is used, it does not hurt).
      self.set_arg(fuzz_args, constants.CMPLOG_LEVEL_FLAG,
                   rand_cmplog_level(self.strategies))

      # Attempt to start the fuzzer.
      fuzz_result = self.run_and_wait(
          additional_args=fuzz_args,
          timeout=max_total_time,
          terminate_before_kill=True,
          terminate_wait_time=self.SIGTERM_WAIT_TIME)

      # Reduce max_total_time by the amount of time the last attempt took.
      max_total_time -= fuzz_result.time_executed

      # Break now only if everything went well. Note that if afl finds a crash
      # from fuzzing (and not in the input) it will exit with a zero return
      # code.
      if fuzz_result.return_code == 0:
        # If afl-fuzz found a crash, copy it to the testcase_file_path.
        self.afl_output.copy_crash_if_needed(self.testcase_file_path)
        break

      # Else the return_code was not 0 so something didn't work out. Try fixing
      # this if afl-fuzz threw an error because it saw a crash, hang or large
      # file in the starting corpus.

      # If there was a crash in the input/corpus, afl-fuzz won't run, so let
      # ClusterFuzz know about this and quit.
      crash_filename = check_error_and_log(self.CRASH_REGEX,
                                           self.CRASH_LOG_MESSAGE)

      if crash_filename:
        crash_path = os.path.join(self.afl_input.input_directory,
                                  crash_filename)

        # Copy this file over so afl can reproduce the crash.
        shutil.copyfile(crash_path, self.testcase_file_path)
        break

      # afl-fuzz won't run if there is a hang in the input.
      hang_filename = check_error_and_log(self.HANG_REGEX,
                                          self.HANG_LOG_MESSAGE)

      if hang_filename:
        # Remove hang from queue and resume fuzzing
        self.afl_output.remove_hang_in_queue(hang_filename)

        # Now that the bad testcase has been removed, let's resume fuzzing so we
        # don't start again from the beginning of the corpus.
        self.set_resume(fuzz_args)

        if hang_filename.startswith('id:000000'):
          num_first_testcase_hangs += 1
          if (num_first_testcase_hangs >
              self.MAX_FIRST_HANGS_WITH_DEFERRED_FORKSERVER):
            logs.log_warn('First testcase hangs when not deferring.')

          elif (num_first_testcase_hangs ==
                self.MAX_FIRST_HANGS_WITH_DEFERRED_FORKSERVER):
            environment.set_value(constants.DONT_DEFER_ENV_VAR, 1)
            print('Instructing AFL not to defer forkserver.\nIf this fixes the '
                  'fuzzer, you should add this to the .options file:\n'
                  '[env]\n'
                  'afl_driver_dont_defer = 1')

        if num_retries - 1 > self.MAX_FUZZ_RETRIES_WITH_STRICT_TIMEOUT:
          skip_hangs = True
          self.set_timeout_arg(fuzz_args, self.MANUAL_TIMEOUT_MILLISECONDS,
                               skip_hangs)

        continue

      # If False: then prepare_retry_if_cpu_error can't solve the issue.
      if self.prepare_retry_if_cpu_error(fuzz_result):
        continue  # Try fuzzing again with the cpu error fixed.

      # If we can't do anything useful about the error, log it and don't try to
      # fuzz again.
      logs.log_error(
          ('Afl exited with a non-zero exitcode: %s. Cannot recover.' %
           fuzz_result.return_code),
          engine_output=fuzz_result.output)

      break

    return fuzz_result

  def prepare_retry_if_cpu_error(self, fuzz_result):
    """AFL will try to bind targets to a particular core for a speed
    improvement. If this isn't possible, then AFL won't run unless
    AFL_NO_AFFINITY=1. One way this can happen is if afl-fuzz leaves zombies
    around that are still bound to a core. Unfortunately AFL sometimes does
    this in persistent mode.
    See https://groups.google.com/forum/#!topic/afl-users/E37s4YDti7o
    """
    if re.search(self.CPU_BIND_ERROR_REGEX, fuzz_result.output) is None:
      return False

    # If we have already tried fixing this error but it is still happening,
    # log it and don't try again.
    current_no_affinity_value = environment.get_value(
        constants.NO_AFFINITY_ENV_VAR)

    if current_no_affinity_value is not None:
      logs.log_warn(('Already tried fixing CPU bind error\n'
                     '$AFL_NO_AFFINITY: %s\n'
                     'Not retrying.') % current_no_affinity_value)

      return False  # return False so this error is considered unhandled.

    # Log that this happened so someone can investigate/remediate the zombies
    # and then try fuzzing again, this time telling AFL not to bind.
    logs.log_error(
        'CPU binding error encountered by afl-fuzz\n'
        'Check bot: %s for zombies\n'
        'Trying again with AFL_NO_AFFINITY=1' % BOT_NAME,
        afl_output=fuzz_result.output)
    environment.set_value(constants.NO_AFFINITY_ENV_VAR, 1)
    return True

  @property
  def afl_input(self):
    """Don't create the object until we need it, since it isn't used for
    reproducing testcases."""
    if self._afl_input is None:
      self._afl_input = AflFuzzInputDirectory(self._input_directory,
                                              self.target_path, self.strategies)

    return self._afl_input

  @property
  def afl_output(self):
    """Don't create the object until we need it, since it isn't used for
    reproducing testcases."""

    if self._afl_output is None:
      self._afl_output = AflFuzzOutputDirectory()
    return self._afl_output

  def fuzz(self):
    """Running fuzzing command. Wrapper around run_afl_fuzz that performs one
    time setup.

    Returns:
      A new_process.ProcessResult.
    """
    # Ensure self.executable_path is afl-fuzz
    self._executable_path = self.afl_fuzz_path

    self.initial_max_total_time = (
        get_fuzz_timeout(
            self.strategies.is_mutations_run, full_timeout=self.timeout) -
        self.AFL_CLEAN_EXIT_TIME - self.SIGTERM_WAIT_TIME)

    assert self.initial_max_total_time > 0

    self._fuzz_args = self.generate_afl_args()

    self.do_offline_mutations()

    return self.run_afl_fuzz(self._fuzz_args)

  def get_file_features(self, input_file_path, showmap_args):
    """Get the features (edge hit counts) of |input_file_path| using
    afl-showmap."""
    # TODO(metzman): Figure out if we should worry about CPU affinity errors
    # here.
    showmap_result = self.run_and_wait(
        additional_args=showmap_args,
        input_data=engine_common.read_data_from_file(input_file_path),
        # TODO(metzman): Set a more reasonable per-file timeout. This is a must
        # to make timeouts smarter for afl-fuzz.
        timeout=self.merge_timeout,
        extra_env={constants.DEBUG_VAR: '1'})

    self.merge_timeout -= showmap_result.time_executed

    # TODO(metzman): Figure out why negative values are accepted by
    # self.run_and_wait.
    if showmap_result.timed_out or self.merge_timeout <= 0:
      return None, True

    # Log an error if showmap didn't write any coverage.
    if not os.path.exists(self.showmap_output_path):
      if not self.showmap_no_output_logged:
        self.showmap_no_output_logged = True
        logs.log_error(('afl-showmap didn\'t output any coverage for '
                        'file {file_path} ({file_size} bytes).\n'
                        'Command: {command}\n'
                        'Return code: {return_code}\n'
                        'Time executed: {time_executed}\n'
                        'Output: {output}').format(
                            file_path=input_file_path,
                            file_size=os.path.getsize(input_file_path),
                            command=showmap_result.command,
                            return_code=showmap_result.return_code,
                            time_executed=showmap_result.time_executed,
                            output=showmap_result.output))
      return None, True

    showmap_output = engine_common.read_data_from_file(self.showmap_output_path)

    features = set()
    for match in re.finditer(self.SHOWMAP_REGEX, showmap_output):
      d = match.groupdict()
      features.add((int(d['guard']), int(d['hit_count'])))
    logs.log('afl-showmap succeedded for '
             'file {file_path}: {features} features'.format(
                 file_path=input_file_path, features=len(features)))
    return frozenset(features), False

  def merge_corpus(self):
    """Merge new testcases into the input corpus."""
    logs.log('Merging corpus.')
    # Don't tell the fuzz target to write its stderr to the same file written
    # to during fuzzing. The target won't write its stderr anywhere.
    try:
      del os.environ[constants.STDERR_FILENAME_ENV_VAR]
    except KeyError:
      pass
    self._executable_path = self.afl_showmap_path
    # Hack around minijail.
    showmap_args = self._fuzz_args
    showmap_args[-1] = '1'
    # Remove arguments that are just for afl-fuzz.
    self.remove_arg(showmap_args, constants.DICT_FLAG)
    self.remove_arg(showmap_args, constants.INPUT_FLAG)
    self.remove_arg(showmap_args, constants.INSTANCE_ID_FLAG)
    self.remove_arg(showmap_args, constants.MOPT_FLAG)
    self.remove_arg(showmap_args, constants.CMPLOG_LEVEL_FLAG)
    self.remove_arg(showmap_args, constants.QUEUE_OLD_STRATEGY_FLAG)
    self.remove_arg(showmap_args, constants.SCHEDULER_FLAG)
    self.remove_arg(showmap_args, constants.CMPLOG_FLAG)

    # Replace -o argument.
    if environment.get_value('USE_MINIJAIL'):
      showmap_output_path = '/' + self.SHOWMAP_FILENAME
    else:
      showmap_output_path = self.showmap_output_path
    idx = self.get_arg_index(showmap_args, constants.OUTPUT_FLAG)
    assert idx != -1
    self.set_arg(showmap_args, constants.OUTPUT_FLAG, showmap_output_path)

    input_dir = self.afl_input.input_directory
    corpus = Corpus()
    input_inodes = set()
    input_filenames = set()
    for file_path in shell.get_files_list(input_dir):
      file_features, timed_out = self.get_file_features(file_path, showmap_args)
      if timed_out:
        logs.log_warn('Timed out in merge while processing initial corpus.')
        return 0

      input_inodes.add(os.stat(file_path).st_ino)
      input_filenames.add(os.path.basename(file_path))
      corpus.associate_features_with_file(file_features, file_path)

    for file_path in list_full_file_paths(self.afl_output.queue):
      # Don't waste time merging copied files.
      inode = os.stat(file_path).st_ino

      # TODO(metzman): Make is_new_testcase capable of checking for hard links
      # and same files.
      # TODO(metzman): Replace this with portable code.
      if (not self.afl_output.is_new_testcase(file_path) or
          inode in input_inodes or  # Is it a hard link?
          # Is it the same file?
          os.path.basename(file_path) in input_filenames):
        continue

      file_features, timed_out = self.get_file_features(file_path, showmap_args)
      if timed_out:
        logs.log_warn('Timed out in merge while processing output.')
        break

      corpus.associate_features_with_file(file_features, file_path)

    # Use destination file as hash of file contents to avoid overwriting
    # different files with the same name that were created from another
    # launcher instance.
    new_units_added = 0
    for src_path in corpus.element_paths:
      # Don't merge files into the initial corpus if they are already there.
      if os.path.dirname(src_path) == input_dir:
        continue
      dest_filename = utils.file_hash(src_path)
      dest_path = os.path.join(input_dir, dest_filename)
      if shell.move(src_path, dest_path):
        new_units_added += 1

    return new_units_added

  def libfuzzerize_corpus(self):
    """Make corpus directories libFuzzer compatible, merge new testcases
    if needed and return the number of new testcases added to corpus.
    """
    # Remove the input directory binding as it isn't needed for merging and
    # may actually break merging if it was temporary and gets deleted.
    if environment.get_value('USE_MINIJAIL'):
      input_directory = self.afl_input.input_directory
      input_bindings = [
          binding for binding in self.chroot.bindings
          if binding.src_path == input_directory
      ]

      assert len(input_bindings) == 1
      input_binding = input_bindings[0]
      self.chroot.bindings.remove(input_binding)

    self.afl_input.restore_if_needed()
    # Number of new units created during fuzzing.
    new_units_generated = self.afl_output.count_new_units(self.afl_output.queue)

    # Number of new units we add to the corpus after merging.
    new_units_added = new_units_generated

    if new_units_generated:
      new_units_added = self.merge_corpus()
      logs.log('Merge completed successfully.')

    # Get corpus size after merge. This removes the duplicate units that were
    # created during this fuzzing session.
    corpus_size = shell.get_directory_file_count(self.afl_input.input_directory)

    return new_units_generated, new_units_added, corpus_size


class AflRunner(AflRunnerCommon, new_process.UnicodeProcessRunner):
  """Afl runner."""

  def __init__(self,
               target_path,
               config,
               testcase_file_path,
               input_directory,
               timeout=None,
               afl_tools_path=None,
               strategy_dict=None):
    super().__init__(target_path, config, testcase_file_path, input_directory,
                     timeout, afl_tools_path, strategy_dict)

    new_process.ProcessRunner.__init__(self, self.afl_fuzz_path)


class MinijailAflRunner(AflRunnerCommon, new_process.UnicodeProcessRunnerMixin,
                        engine_common.MinijailEngineFuzzerRunner):
  """Minijail AFL runner."""

  def __init__(self,
               chroot,
               target_path,
               config,
               testcase_file_path,
               input_directory,
               timeout=None,
               afl_tools_path=None,
               strategy_dict=None):
    super().__init__(target_path, config, testcase_file_path, input_directory,
                     timeout, afl_tools_path, strategy_dict)

    minijail.MinijailProcessRunner.__init__(self, chroot, self.afl_fuzz_path)

  def _get_or_create_chroot_binding(self, corpus_directory):
    """Return chroot relative paths for the given corpus directories.

    Args:
      corpus_directories: A list of host corpus directories.

    Returns:
      A list of chroot relative paths.
    """
    chroot_rel_dir = os.path.relpath(corpus_directory, self.chroot.directory)
    if not chroot_rel_dir.startswith(os.pardir):
      # Already in chroot.
      return '/' + chroot_rel_dir

    binding = self.chroot.get_binding(corpus_directory)
    if binding:
      return binding.dest_path

    dest_path = '/' + os.path.basename(corpus_directory)
    self.chroot.add_binding(
        minijail.ChrootBinding(corpus_directory, dest_path, True))

    return dest_path

  def run_single_testcase(self, testcase_path):
    with self._chroot_testcase(testcase_path) as chroot_testcase_path:
      return super().run_single_testcase(chroot_testcase_path)

  def generate_afl_args(self,
                        afl_input=None,
                        afl_output=None,
                        mem_limit=constants.MAX_MEMORY_LIMIT):
    """Overriden generate_afl_args."""
    if afl_input:
      minijail_afl_input = self._get_or_create_chroot_binding(afl_input)
    else:
      minijail_afl_input = self._get_or_create_chroot_binding(
          self.afl_input.input_directory)

    if afl_output:
      minijail_afl_output = self._get_or_create_chroot_binding(afl_output)
    else:
      minijail_afl_output = self._get_or_create_chroot_binding(
          self.afl_output.output_directory)

    return super().generate_afl_args(minijail_afl_input, minijail_afl_output,
                                     mem_limit)

  @property
  def stderr_file_path(self):
    """Overriden stderr_file_path."""
    return os.path.join(self.chroot.directory, STDERR_FILENAME)

  def set_environment_variables(self):
    """Overridden set_environment_variables."""
    super().set_environment_variables()
    environment.set_value(constants.STDERR_FILENAME_ENV_VAR,
                          '/' + STDERR_FILENAME)


class CorpusElement(object):
  """An element (file) in a corpus."""

  def __init__(self, path):
    self.path = path
    self.size = os.path.getsize(self.path)


class Corpus(object):
  """A minimal set of input files (elements) for a fuzz target."""

  def __init__(self):
    self.features_and_elements = {}

  @property
  def element_paths(self):
    """Returns the filepaths of all elements in the corpus."""
    return set(
        element.path for element in six.itervalues(self.features_and_elements))

  def _associate_feature_with_element(self, feature, element):
    """Associate a feature with an element if the element is the smallest for
    the feature."""
    if feature not in self.features_and_elements:
      self.features_and_elements[feature] = element
      return

    # Feature already has an associated element.
    incumbent_element = self.features_and_elements[feature]
    if incumbent_element.size > element.size:
      self.features_and_elements[feature] = element

  def associate_features_with_file(self, features, path):
    """Associate features with a file when the file is the smallest for the
    features."""
    element = CorpusElement(path)
    for feature in features:
      self._associate_feature_with_element(feature, element)


def _verify_system_config():
  """Verifies system settings required for AFL."""

  def _check_core_pattern_file():
    """Verifies that core pattern file content is set to 'core'."""
    if not os.path.exists(constants.CORE_PATTERN_FILE_PATH):
      return False

    return open(constants.CORE_PATTERN_FILE_PATH).read().strip() == 'core'

  if _check_core_pattern_file():
    return

  return_code = subprocess.call(
      'sudo -n bash -c "echo core > {path}"'.format(
          path=constants.CORE_PATTERN_FILE_PATH),
      shell=True)
  if return_code or not _check_core_pattern_file():
    logs.log_fatal_and_exit(
        'Failed to set {path}. AFL needs {path} to be set to core.'.format(
            path=constants.CORE_PATTERN_FILE_PATH))


def load_testcase_if_exists(fuzzer_runner, testcase_file_path):
  """Loads a crash testcase if it exists."""
  # To ensure that we can run the fuzzer.
  os.chmod(fuzzer_runner.executable_path, stat.S_IRWXU | stat.S_IRGRP
           | stat.S_IXGRP)

  fuzzer_runner.run_single_testcase(testcase_file_path)
  print(fuzzer_runner.fuzzer_stderr)
  return True


def set_additional_sanitizer_options_for_afl_fuzz():
  """Set *SAN_OPTIONS to afl's liking.

  If ASAN_OPTIONS or MSAN_OPTION is set, they must contain certain options or
  afl-fuzz will refuse to fuzz. See check_asan_opts() in afl-fuzz.c in afl for
  more details.
  """
  # We need to check if ASAN_OPTIONS and/or MSAN_OPTIONS contain symbolize=0
  # because ClusterFuzz sets all sanitizers options equal to an empty string
  # before adding symbolize=0 to *either* ASAN_OPTIONS or MSAN_OPTIONS. Because
  # they will both be set but one will be empty, afl will think the empty one is
  # incorrect and quit if we don't do this.
  required_sanitizer_options = {
      'ASAN_OPTIONS': {
          'symbolize': 0,
          'abort_on_error': 1
      },
      'MSAN_OPTIONS': {
          'symbolize': 0,
          'exit_code': 86
      },
  }

  for options_env_var, option_values in six.iteritems(
      required_sanitizer_options):
    # If os.environ[options_env_var] is an empty string, afl will refuse to run,
    # because we haven't set the right options. Thus only continue if it does
    # not exist.
    if options_env_var not in os.environ:
      continue

    options_env_value = environment.get_memory_tool_options(options_env_var)
    options_env_value.update(option_values)
    environment.set_memory_tool_options(options_env_var, options_env_value)


def remove_path(path):
  """Remove |path| if it exists. Similar to running rm -rf |path|."""
  # Remove links to files and files.
  if os.path.isfile(path):
    os.remove(path)
  elif os.path.isdir(path):
    shutil.rmtree(path)
  # Else path doesn't exist. Do nothing.


def list_full_file_paths(directory):
  """List the absolute paths of files in |directory|."""
  directory_absolute_path = os.path.abspath(directory)
  paths = []
  for relative_path in os.listdir(directory):
    absolute_path = os.path.join(directory_absolute_path, relative_path)
    if os.path.isfile(absolute_path):  # Only return paths to files.
      paths.append(absolute_path)
  return paths


def get_first_stacktrace(stderr_data):
  """If |stderr_data| contains stack traces, only returns the first one.
  Otherwise returns the entire string."""

  # Use question mark after .+ for non-greedy, otherwise it will match more
  # than one stack trace.
  sanitizer_stacktrace_regex = r'ERROR: [A-z]+Sanitizer: .*\n(.|\n)+?ABORTING'
  match = re.search(sanitizer_stacktrace_regex, stderr_data)

  # If we can't find the first stacktrace, return the whole thing.
  if match is None:
    return stderr_data

  return stderr_data[:match.end()]


# TODO(mbarbella): After deleting the non-engine AFL code, remove this
# function and replace it with a simple check based on the timeout set in
# AFLEngine.fuzz. Mutations should also be moved to fuzz.
def get_fuzz_timeout(is_mutations_run, full_timeout=None):
  """Get the maximum amount of time that should be spent fuzzing."""
  fuzz_timeout = full_timeout

  # TODO(mbarbella): Delete this check once non-engine support is removed. The
  # engine pipeline always specifies a timeout.
  if fuzz_timeout is None:
    fuzz_timeout = engine_common.get_hard_timeout() - POSTPROCESSING_TIMEOUT

  # Subtract mutations timeout from the total timeout.
  if is_mutations_run:
    fuzz_timeout -= engine_common.get_new_testcase_mutations_timeout()

  # Subtract the merge timeout from the fuzz timeout in the non-engine case.
  if not full_timeout:
    fuzz_timeout -= engine_common.get_merge_timeout(DEFAULT_MERGE_TIMEOUT)

  assert fuzz_timeout > 0
  return fuzz_timeout


def prepare_runner(fuzzer_path,
                   config,
                   testcase_file_path,
                   input_directory,
                   timeout=None,
                   strategy_dict=None):
  """Common initialization code shared by the new pipeline and main."""
  # Set up temp dir.
  engine_common.recreate_directory(fuzzer_utils.get_temp_dir())

  if environment.get_value('USE_MINIJAIL'):
    # Set up chroot and runner.
    minijail_chroot = minijail.MinijailChroot(
        base_dir=fuzzer_utils.get_temp_dir())

    build_dir = environment.get_value('BUILD_DIR')

    # While it's possible for dynamic binaries to run without this, they need to
    # be accessible for symbolization etc. For simplicity we bind BUILD_DIR to
    # the same location within the chroot, which leaks the directory structure
    # of CF but this shouldn't be a big deal.
    minijail_chroot.add_binding(
        minijail.ChrootBinding(build_dir, build_dir, False))

    # AFL expects various things in /bin.
    minijail_chroot.add_binding(minijail.ChrootBinding('/bin', '/bin', False))

    # And /usr/bin.
    minijail_chroot.add_binding(
        minijail.ChrootBinding('/usr/bin', '/usr/bin', False))

    # Also bind the build dir to /out to make it easier to hardcode references
    # to data files.
    minijail_chroot.add_binding(
        minijail.ChrootBinding(build_dir, '/out', False))

    # map /proc/self/fd -> /dev/fd
    os.symlink('/proc/self/fd',
               os.path.join(minijail_chroot.directory, 'dev', 'fd'))

    runner = MinijailAflRunner(
        minijail_chroot,
        fuzzer_path,
        config,
        testcase_file_path,
        input_directory,
        timeout=timeout,
        strategy_dict=strategy_dict)

  else:
    runner = AflRunner(
        fuzzer_path,
        config,
        testcase_file_path,
        input_directory,
        timeout=timeout,
        strategy_dict=strategy_dict)

  # Make sure afl won't exit because of bad sanitizer options.
  set_additional_sanitizer_options_for_afl_fuzz()

  # Add *SAN_OPTIONS overrides from .options file.
  engine_common.process_sanitizer_options_overrides(fuzzer_path)

  return runner


def rand_schedule(scheduler_probs):
  """Returns a random queue scheduler."""
  schedule = 'fast'
  rnd = engine_common.get_probability()
  for schedule_opt, prob in scheduler_probs:
    if rnd > prob:
      rnd -= prob
    else:
      schedule = schedule_opt
      break
  return schedule


def rand_cmplog_level(strategies):
  """Returns a random CMPLOG intensity level."""
  cmplog_level = '2'
  rnd = engine_common.get_probability()
  for cmplog_level_opt, prob in strategies.CMPLOG_LEVEL_PROBS:
    if rnd > prob:
      rnd -= prob
    else:
      cmplog_level = cmplog_level_opt
      break
  if engine_common.decide_with_probability(strategies.CMPLOG_ARITH_PROB):
    cmplog_level += constants.CMPLOG_ARITH
  if engine_common.decide_with_probability(strategies.CMPLOG_TRANS_PROB):
    cmplog_level += constants.CMPLOG_TRANS
  return cmplog_level


# pylint: disable=too-many-function-args
def main(argv):
  """Run afl as specified by argv."""
  atexit.register(fuzzer_utils.cleanup)

  # Initialize variables.
  testcase_file_path = argv[1]
  target_name = environment.get_value('FUZZ_TARGET')
  input_directory = environment.get_value('FUZZ_CORPUS_DIR')

  # FIXME: Remove this once AFL is migrated to the new engine impl and runs in
  # same python process.
  logs.configure('run_fuzzer')
  _verify_system_config()
  profiler.start_if_needed('afl_launcher')

  build_directory = environment.get_value('BUILD_DIR')
  fuzzer_path = engine_common.find_fuzzer_path(build_directory, target_name)
  if not fuzzer_path:
    return

  # Install signal handler.
  signal.signal(signal.SIGTERM, engine_common.signal_term_handler)

  config = AflConfig.from_target_path(fuzzer_path)
  runner = prepare_runner(fuzzer_path, config, testcase_file_path,
                          input_directory)

  # If we don't have a corpus, then that means this is not a fuzzing run.
  if not input_directory:
    load_testcase_if_exists(runner, testcase_file_path)
    return

  # Execute afl-fuzz on the fuzzing target.
  fuzz_result = runner.fuzz()

  command = fuzz_result.command
  if environment.get_value('USE_MINIJAIL'):
    command = engine_common.strip_minijail_command(command,
                                                   runner.afl_fuzz_path)
  # Print info for the fuzzer logs.
  print(engine_common.get_log_header(command, fuzz_result.time_executed))

  print(fuzz_result.output)

  if fuzz_result.return_code:
    # If AFL returned a non-zero return code quit now without getting stats,
    # since they would be meaningless.
    print(runner.fuzzer_stderr)
    return

  stats_getter = stats.StatsGetter(runner.afl_output.stats_path,
                                   config.dict_path)
  try:
    new_units_generated, new_units_added, corpus_size = (
        runner.libfuzzerize_corpus())
    stats_getter.set_stats(fuzz_result.time_executed, new_units_generated,
                           new_units_added, corpus_size, runner.strategies,
                           runner.fuzzer_stderr, fuzz_result.output)

    engine_common.dump_big_query_data(stats_getter.stats, testcase_file_path,
                                      command)

  finally:
    print(runner.fuzzer_stderr)

  # Record the stats to make them easily searchable in stackdriver.
  if new_units_added:
    logs.log(
        'New units added to corpus: %d.' % new_units_added,
        stats=stats_getter.stats)
  else:
    logs.log('No new units found.', stats=stats_getter.stats)


if __name__ == '__main__':
  main(sys.argv)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for ml_train."""

import unittest

import flask
import webtest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.cron import ml_train


@test_utils.with_cloud_emulators('datastore')
class HandlerTest(unittest.TestCase):
  """Tests for Handler."""

  def setUp(self):
    test_helpers.patch_environ(self)
    flaskapp = flask.Flask('testflask')
    flaskapp.add_url_rule(
        '/schedule-ml-train-tasks',
        view_func=ml_train.Handler.as_view('/schedule-ml-train-tasks'))
    self.app = webtest.TestApp(flaskapp)

    test_helpers.patch(self, [
        'clusterfuzz._internal.base.tasks.add_task',
        'handlers.base_handler.Handler.is_cron',
        'clusterfuzz._internal.metrics.logs.log_error'
    ])

    # Create fake jobs.
    data_types.Job(
        name='libfuzzer_asan',
        environment_string='ML_MODELS_TO_USE = rnn_generator').put()
    data_types.Job(name='libfuzzer_msan', environment_string='').put()
    data_types.Job(name='afl_asan', environment_string='').put()

    data_types.Job(
        name='libfuzzer_asan_invalid',
        environment_string='ML_MODELS_TO_USE = invalid_model\n').put()

    # Create fake fuzzers.
    data_types.Fuzzer(name='afl', jobs=['afl_asan']).put()

    # Create fake child fuzzers.
    data_types.FuzzTarget(
        engine='libFuzzer', binary='fake_fuzzer', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='libFuzzer_fake_fuzzer', job='libfuzzer_asan').put()
    data_types.FuzzTarget(
        engine='afl', binary='fake_fuzzer', project='test-project').put()
    data_types.FuzzTargetJob(
        fuzz_target_name='afl_fake_fuzzer', job='afl_asan').put()

  def test_add_tasks(self):
    """Tests adding single and multiple tasks."""
    self.app.get('/schedule-ml-train-tasks')
    self.mock.add_task.assert_any_call(
        'train_rnn_generator',
        'libFuzzer_fake_fuzzer',
        'libfuzzer_asan',
        queue='ml-jobs-linux')

    # Ensure that we logged an error for the invalid model.
    self.mock.log_error.assert_called_once_with(
        'Invalid ML model invalid_model for job libfuzzer_asan_invalid.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Minimizer helper functions."""

import os
import shlex
import subprocess

from clusterfuzz._internal.bot.tokenizer.antlr_tokenizer import AntlrTokenizer
from clusterfuzz._internal.bot.tokenizer.grammars.HTMLLexer import HTMLLexer

from . import errors

# TODO(mbarbella): Improve configuration of the test function.
attempts = 1
test_command = None

# This list of markers is a copy of code in ClusterFuzz.
# This is kept here to keep the code standalone.
STACKTRACE_TOOL_MARKERS = [
    ' runtime error: ',
    'AddressSanitizer',
    'ASAN:',
    'CFI: Most likely a control flow integrity violation;',
    'ERROR: libFuzzer',
    'KASAN:',
    'LeakSanitizer',
    'MemorySanitizer',
    'ThreadSanitizer',
    'UndefinedBehaviorSanitizer',
    'UndefinedSanitizer',
]
STACKTRACE_END_MARKERS = [
    'ABORTING',
    'END MEMORY TOOL REPORT',
    'End of process memory map.',
    'END_KASAN_OUTPUT',
    'SUMMARY:',
    'Shadow byte and word',
    '[end of stack trace]',
    '\nExiting',
    'minidump has been written',
]
CHECK_FAILURE_MARKERS = [
    'Check failed:',
    'Device rebooted',
    'Fatal error in',
    'FATAL EXCEPTION',
    'JNI DETECTED ERROR IN APPLICATION:',
    'Sanitizer CHECK failed:',
]


def get_size_string(size):
  """Return string representation for size."""
  if size < 1 << 10:
    return '%d B' % size
  if size < 1 << 20:
    return '%d KB' % (size >> 10)
  if size < 1 << 30:
    return '%d MB' % (size >> 20)

  return '%d GB' % (size >> 30)


def has_marker(stacktrace, marker_list):
  """Return true if the stacktrace has atleast one marker
  in the marker list."""
  for marker in marker_list:
    if marker in stacktrace:
      return True

  return False


def set_test_command(new_test_command):
  """Set the command used for testing."""
  global test_command
  test_command = shlex.split(new_test_command)


def set_test_attempts(new_attempts):
  """Set the number of times to attempt the test."""
  global attempts
  attempts = new_attempts


def test(test_path):
  """Wrapper function to verify that a test does not fail for multiple runs."""
  for _ in range(attempts):
    if not single_test_run(test_path):
      return False
  return True


def single_test_run(test_path):
  """Hacky test function that checks for certain common errors."""
  if not test_command:
    raise errors.NoCommandError

  args = test_command + [test_path]
  try:
    console_output = subprocess.check_output(args, stderr=subprocess.STDOUT)
  except subprocess.CalledProcessError as error:
    console_output = error.output

  # If we meet one of these conditions, assume we crashed.
  if ((has_marker(console_output, STACKTRACE_TOOL_MARKERS) and
       has_marker(console_output, STACKTRACE_END_MARKERS)) or
      has_marker(console_output, CHECK_FAILURE_MARKERS)):
    print('Crashed, current test size %s.' % (get_size_string(
        os.path.getsize(test_path))))
    return False

  # No crash, test passed.
  print('Not crashed, current test size %s.' % (get_size_string(
      os.path.getsize(test_path))))
  return True


def tokenize(data):
  """HTML tokenizer."""
  return AntlrTokenizer(HTMLLexer).tokenize(data)


def token_combiner(tokens):
  """Dummy token combiner."""
  return ''.join(tokens)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tasks host."""

import datetime

from google.protobuf import wrappers_pb2
import grpc
import six

from clusterfuzz._internal.bot import testcase_manager
from clusterfuzz._internal.bot.tasks import corpus_pruning_task
from clusterfuzz._internal.bot.untrusted_runner import file_host
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz.fuzz import engine

from . import host


def _fuzz_target_to_proto(fuzz_target):
  """Convert fuzz_target to protobuf."""
  return untrusted_runner_pb2.FuzzTarget(
      engine=fuzz_target.engine,
      project=fuzz_target.project,
      binary=fuzz_target.binary,
  )


def do_corpus_pruning(context, last_execution_failed, revision):
  """Do corpus pruning on untrusted worker."""
  cross_pollinate_fuzzers = [
      untrusted_runner_pb2.CrossPollinateFuzzer(
          fuzz_target=_fuzz_target_to_proto(cpf.fuzz_target),
          backup_bucket_name=cpf.backup_bucket_name,
          corpus_engine_name=cpf.corpus_engine_name,
      ) for cpf in context.cross_pollinate_fuzzers
  ]

  request = untrusted_runner_pb2.PruneCorpusRequest(
      fuzz_target=_fuzz_target_to_proto(context.fuzz_target),
      cross_pollinate_fuzzers=cross_pollinate_fuzzers,
      last_execution_failed=last_execution_failed,
      revision=revision)

  response = host.stub().PruneCorpus(request)

  project_qualified_name = context.fuzz_target.project_qualified_name()
  today_date = datetime.datetime.utcnow().date()
  coverage_info = data_types.CoverageInformation(
      fuzzer=project_qualified_name, date=today_date)

  # Intentionally skip edge and function coverage values as those would come
  # from fuzzer coverage cron task (see src/go/server/cron/coverage.go).
  coverage_info.corpus_size_units = response.coverage_info.corpus_size_units
  coverage_info.corpus_size_bytes = response.coverage_info.corpus_size_bytes
  coverage_info.corpus_location = response.coverage_info.corpus_location
  coverage_info.corpus_backup_location = (
      response.coverage_info.corpus_backup_location)
  coverage_info.quarantine_size_units = (
      response.coverage_info.quarantine_size_units)
  coverage_info.quarantine_size_bytes = (
      response.coverage_info.quarantine_size_bytes)
  coverage_info.quarantine_location = response.coverage_info.quarantine_location

  crashes = [
      corpus_pruning_task.CorpusCrash(
          crash_state=crash.crash_state,
          crash_type=crash.crash_type,
          crash_address=crash.crash_address,
          crash_stacktrace=crash.crash_stacktrace,
          unit_path=crash.unit_path,
          security_flag=crash.security_flag,
      ) for crash in response.crashes
  ]

  result_stats = response.cross_pollination_stats
  pollination_stats = corpus_pruning_task.CrossPollinationStats(
      project_qualified_name=result_stats.project_qualified_name,
      method=result_stats.method,
      sources=result_stats.sources,
      tags=result_stats.tags,
      initial_corpus_size=result_stats.initial_corpus_size,
      corpus_size=result_stats.corpus_size,
      initial_edge_coverage=result_stats.initial_edge_coverage,
      edge_coverage=result_stats.edge_coverage,
      initial_feature_coverage=result_stats.initial_feature_coverage,
      feature_coverage=result_stats.feature_coverage)

  return corpus_pruning_task.CorpusPruningResult(
      coverage_info=coverage_info,
      crashes=crashes,
      fuzzer_binary_name=response.fuzzer_binary_name,
      revision=response.revision,
      cross_pollination_stats=pollination_stats)


def process_testcase(engine_name, tool_name, target_name, arguments,
                     testcase_path, output_path, timeout):
  """Process testcase on untrusted worker."""
  if tool_name == 'minimize':
    operation = untrusted_runner_pb2.ProcessTestcaseRequest.MINIMIZE
  else:
    operation = untrusted_runner_pb2.ProcessTestcaseRequest.CLEANSE

  rebased_testcase_path = file_host.rebase_to_worker_root(testcase_path)
  file_host.copy_file_to_worker(testcase_path, rebased_testcase_path)

  request = untrusted_runner_pb2.ProcessTestcaseRequest(
      engine=engine_name,
      operation=operation,
      target_name=target_name,
      arguments=arguments,
      testcase_path=file_host.rebase_to_worker_root(testcase_path),
      output_path=file_host.rebase_to_worker_root(output_path),
      timeout=timeout)

  response = host.stub().ProcessTestcase(request)

  rebased_output_path = file_host.rebase_to_worker_root(output_path)
  file_host.copy_file_from_worker(rebased_output_path, output_path)

  return engine.ReproduceResult(
      list(response.command), response.return_code, response.time_executed,
      response.output)


def _unpack_values(values):
  """Unpack protobuf values."""
  unpacked = {}
  for key, packed_value in six.iteritems(values):
    if packed_value.Is(wrappers_pb2.DoubleValue.DESCRIPTOR):
      value = wrappers_pb2.DoubleValue()
    elif packed_value.Is(wrappers_pb2.Int64Value.DESCRIPTOR):
      value = wrappers_pb2.Int64Value()
    elif packed_value.Is(wrappers_pb2.StringValue.DESCRIPTOR):
      value = wrappers_pb2.StringValue()
    else:
      raise ValueError('Unknown stat type for ' + key)

    packed_value.Unpack(value)
    unpacked[key] = value.value

  return unpacked


def engine_fuzz(engine_impl, target_name, sync_corpus_directory,
                testcase_directory):
  """Run engine fuzzer on untrusted worker."""
  request = untrusted_runner_pb2.EngineFuzzRequest(
      engine=engine_impl.name,
      target_name=target_name,
      sync_corpus_directory=file_host.rebase_to_worker_root(
          sync_corpus_directory),
      testcase_directory=file_host.rebase_to_worker_root(testcase_directory))

  response = host.stub().EngineFuzz(request)
  crashes = [
      engine.Crash(
          input_path=file_host.rebase_to_host_root(crash.input_path),
          stacktrace=crash.stacktrace,
          reproduce_args=crash.reproduce_args,
          crash_time=crash.crash_time) for crash in response.crashes
  ]

  unpacked_stats = _unpack_values(response.stats)
  unpacked_strategies = _unpack_values(response.strategies)

  result = engine.FuzzResult(
      logs=response.logs,
      command=list(response.command),
      crashes=crashes,
      stats=unpacked_stats,
      time_executed=response.time_executed)

  file_host.pull_testcases_from_worker()
  return result, dict(response.fuzzer_metadata), unpacked_strategies


def engine_reproduce(engine_impl, target_name, testcase_path, arguments,
                     timeout):
  """Run engine reproduce on untrusted worker."""
  rebased_testcase_path = file_host.rebase_to_worker_root(testcase_path)
  file_host.copy_file_to_worker(testcase_path, rebased_testcase_path)

  request = untrusted_runner_pb2.EngineReproduceRequest(
      engine=engine_impl.name,
      target_name=target_name,
      testcase_path=rebased_testcase_path,
      arguments=arguments,
      timeout=timeout)

  try:
    response = host.stub().EngineReproduce(request)
  except grpc.RpcError as e:
    if 'TargetNotFoundError' in repr(e):
      # Resurface the right exception.
      raise testcase_manager.TargetNotFoundError('Failed to find target ' +
                                                 target_name)
    raise

  return engine.ReproduceResult(
      list(response.command), response.return_code, response.time_executed,
      response.output)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Android kernel util functions."""

import os
import re

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.platforms.android import constants
from clusterfuzz._internal.platforms.android import settings
from clusterfuzz._internal.platforms.android import symbols_downloader
from clusterfuzz._internal.system import environment

# Linux version 3.18.0-g(8de8e79)-ab(1234567) where 8de8e79 is the hash and
# 1234567 optional is the kernel build id.
LINUX_VERSION_REGEX = re.compile(
    r'Linux version .+-g([0-9a-f]+)[\s\-](ab([0-9a-f]+)\s)')

# Similar to source_mapper.STACK_FRAME_PATH_LINE_REGEX but also matches column
# number.
LLVM_SYMBOLIZER_STACK_FRAME_PATH_LINE_REGEX = re.compile(
    r'(?<=\[|\(|\s)([a-zA-Z/.][^\s]*?)\s*(:|@)\s*(\d+)(?=\]$|\)$|:\d+$|$)'
    r'(:\d+)?')


def _get_clean_kernel_path(path):
  """Sometimes kernel paths start with
  /buildbot/src/partner-android/BRANCH/private/PROJ.  We want to remove all
  of this."""
  # Remove the stuff before 'private', since it is build-server dependent
  # and not part of the final URL.
  if '/private/' in path:
    path_parts = path.split('/')
    path_parts = path_parts[path_parts.index('private') + 2:]
    return '/'.join(path_parts)

  return path


def get_kernel_stack_frame_link(stack_frame, kernel_prefix, kernel_hash):
  """Add source links data to kernel stack frames."""
  match = LLVM_SYMBOLIZER_STACK_FRAME_PATH_LINE_REGEX.search(stack_frame)
  if not match:
    # If this stack frame does not contain a path and line, bail out.
    return stack_frame

  path = _get_clean_kernel_path(match.group(1))
  line = match.group(3)
  kernel_prefix = utils.strip_from_left(kernel_prefix, 'kernel/private/')
  display_path = f'{kernel_prefix}/{path}:{line}'

  # If we have a column number, lets add it to the display path.
  if match.group(4):
    display_path += match.group(4)

  kernel_url_info = (r'http://go/pakernel/{prefix}/+/{hash}/{path}#{line};'
                     r'{display_path};').format(
                         prefix=kernel_prefix,
                         hash=kernel_hash,
                         path=path,
                         line=line,
                         display_path=display_path)

  link_added_stack_frame = LLVM_SYMBOLIZER_STACK_FRAME_PATH_LINE_REGEX.sub(
      kernel_url_info, stack_frame)

  return link_added_stack_frame


def _get_prefix_and_full_hash(repo_data, kernel_partial_hash):
  """Find the prefix and full hash in the repo_data based on the partial."""
  kernel_partial_hash_lookup = 'u\'%s' % kernel_partial_hash
  for line in repo_data.splitlines():
    if kernel_partial_hash_lookup in line:
      prefix, full_hash = line.split(' ', 1)
      return prefix, full_hash.strip('u\'')

  return None, None


def _get_repo_prop_data(build_id, target):
  """Downloads repo.prop and returns the data based on build_id and target."""
  symbols_directory = os.path.join(
      environment.get_value('SYMBOLS_DIR'), 'kernel')
  repro_filename = symbols_downloader.get_repo_prop_archive_filename(
      build_id, target)

  # Grab repo.prop, it is not on the device nor in the build_dir.
  symbols_downloader.download_kernel_repo_prop_if_needed(symbols_directory)
  local_repo_path = utils.find_binary_path(symbols_directory, repro_filename)

  if local_repo_path and os.path.exists(local_repo_path):
    return utils.read_data_from_file(local_repo_path, eval_data=False).decode()

  return None


def get_kernel_prefix_and_full_hash():
  """Download repo.prop and return the full hash and prefix."""

  kernel_partial_hash, build_id = get_kernel_hash_and_build_id()
  target = get_kernel_name()
  if not build_id or not target:
    logs.log_error('Could not get kernel parameters, exiting.')
    return None

  android_kernel_repo_data = _get_repo_prop_data(build_id, target)
  if android_kernel_repo_data:
    return _get_prefix_and_full_hash(android_kernel_repo_data,
                                     kernel_partial_hash)

  return None, None


def get_kernel_name():
  """Returns the kernel name for the device, since some kernels are shared."""
  product_name = settings.get_product_name()
  build_product = settings.get_build_product()

  # Strip _kasan off of the end as we will add it later if needed.
  utils.strip_from_right(product_name, '_kasan')

  # Some devices have a different kernel name than product_name, if so use the
  # kernel name.
  return constants.PRODUCT_TO_KERNEL.get(build_product, product_name)


def get_kernel_hash_and_build_id():
  """Returns the (partial_hash, build_id) of the kernel."""
  version_string = settings.get_kernel_version_string()
  match = re.match(LINUX_VERSION_REGEX, version_string)
  if match:
    return match.group(2), match.group(3)

  return None



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Minimizer that attempts to use specialized minimizers on different parts of
   an HTML test case."""

import functools

from clusterfuzz._internal.bot.tokenizer.antlr_tokenizer import AntlrTokenizer
from clusterfuzz._internal.bot.tokenizer.grammars.HTMLLexer import HTMLLexer
from clusterfuzz._internal.bot.tokenizer.grammars.JavaScriptLexer import \
    JavaScriptLexer

from . import chunk_minimizer
from . import delta_minimizer
from . import js_minimizer
from . import minimizer
from . import utils

SCRIPT_START_STRING = b'<script'
SCRIPT_END_STRING = b'</script>'


class HTMLMinimizer(minimizer.Minimizer):  # pylint:disable=abstract-method
  """Specialized HTML minimizer.

     Note that this will not work properly with normal tokenizers. It simply
     acts as a wrapper around other minimizers and passes pieces of the HTML
     file to those."""

  class Token(object):
    """Helper class to represent a single token."""
    TYPE_HTML = 0
    TYPE_SCRIPT = 1

    def __init__(self, data, token_type):
      self.data = data
      self.token_type = token_type

  class TokenizerState(object):
    """Enum for tokenizer states."""
    SEARCHING_FOR_SCRIPT = 0
    SEARCHING_FOR_TAG_END = 1
    SEARCHING_FOR_CLOSE_SCRIPT = 2

  HTMLTOKENIZER = AntlrTokenizer(HTMLLexer).tokenize
  JSTOKENIZER = AntlrTokenizer(JavaScriptLexer).tokenize

  TOKENIZER_MAP = {
      Token.TYPE_HTML: [HTMLTOKENIZER, HTMLTOKENIZER, HTMLTOKENIZER],
      Token.TYPE_SCRIPT: [JSTOKENIZER, JSTOKENIZER],
  }

  CHUNK_SIZES = [
      [400, 100, 20, 5],
      [400, 100, 20, 5, 2],
      [400, 100, 20, 5, 1],
  ]

  def __init__(self, test_function, *args, **kwargs):
    # The HTML minimizer will not be used directly. Instead, preserve its
    # arguments and pass them along when creating subminimizers.
    super(HTMLMinimizer, self).__init__(lambda: False)

    assert not args, 'Positional arguments not supported.'
    assert 'tokenizer' not in kwargs, 'Custom tokenizers not supported.'
    assert 'token_combiner' not in kwargs, 'Custom tokenizers not supported.'

    self.test_function = test_function
    self.kwargs = kwargs

  def minimize(self, data):
    """Wrapper to perform common tasks and call |_execute|."""
    # Do an initial line-by-line minimization to filter out noise.
    line_minimizer = delta_minimizer.DeltaMinimizer(self.test_function,
                                                    **self.kwargs)
    # Do two line minimizations to make up for the fact that minimzations on
    # bots don't always minimize as much as they can.
    for _ in range(2):
      data = line_minimizer.minimize(data)

    tokens = self.get_tokens_and_metadata(data)
    for index, token in enumerate(tokens):
      current_tokenizers = self.TOKENIZER_MAP[token.token_type]
      prefix = self.combine_tokens(tokens[:index])
      suffix = self.combine_tokens(tokens[index + 1:])
      token_combiner = functools.partial(
          self.combine_worker_tokens, prefix=prefix, suffix=suffix)

      for level, current_tokenizer in enumerate(current_tokenizers):
        # We need to preserve the parts of the test case that are not currently
        # being minimized. Create a special token combiner that adds these
        # portions of the test to the combined tokens.

        if token.token_type == HTMLMinimizer.Token.TYPE_HTML:
          current_minimizer = chunk_minimizer.ChunkMinimizer(
              self.test_function,
              chunk_sizes=HTMLMinimizer.CHUNK_SIZES[level],
              token_combiner=token_combiner,
              tokenizer=current_tokenizer,
              **self.kwargs)
        else:
          current_minimizer = js_minimizer.JSMinimizer(
              self.test_function,
              token_combiner=token_combiner,
              tokenizer=current_tokenizer,
              **self.kwargs)

        result_data = current_minimizer.minimize(token.data)
        start = len(prefix)
        end = len(result_data) - len(suffix)
        token.data = result_data[start:end]

    # TODO(mbarbella): Remove this once other minimizers are improved.
    # Do a final line-by-line minimization pass.
    data = self.combine_tokens(tokens)
    return line_minimizer.minimize(data)

  @staticmethod
  def get_tokens_and_metadata(data):
    """Get the token list with associated metadata."""
    tokens = []
    state = HTMLMinimizer.TokenizerState.SEARCHING_FOR_SCRIPT
    current_token_start = 0
    current_token_type = HTMLMinimizer.Token.TYPE_HTML
    index = 0

    while 0 <= index < len(data):
      if state == HTMLMinimizer.TokenizerState.SEARCHING_FOR_SCRIPT:
        # In this case, we are simply searching for the next script tag.
        index = data.find(SCRIPT_START_STRING, index)
        state = HTMLMinimizer.TokenizerState.SEARCHING_FOR_TAG_END

      elif state == HTMLMinimizer.TokenizerState.SEARCHING_FOR_TAG_END:
        # Make sure that this really looks like a script tag.
        next_newline = data.find(b'\n', index)
        tag_end = data.find(b'>', index)
        if 0 <= tag_end < next_newline or next_newline < 0 <= tag_end:
          # The end of the script tag is before the next newline, so it should
          # be safe to attempt to split this.
          index = tag_end + 1
          token = HTMLMinimizer.Token(data[current_token_start:index],
                                      current_token_type)
          tokens.append(token)

          # Update state.
          current_token_type = HTMLMinimizer.Token.TYPE_SCRIPT
          current_token_start = index
          state = HTMLMinimizer.TokenizerState.SEARCHING_FOR_CLOSE_SCRIPT
        else:
          # We found a newline before the end of tag or did not find the end
          # of the tag, so something seems wrong. Skip this one.
          index += len(SCRIPT_START_STRING)

      elif state == HTMLMinimizer.TokenizerState.SEARCHING_FOR_CLOSE_SCRIPT:
        # Simply look for the end of this script.
        index = data.find(SCRIPT_END_STRING, index)
        if index < 0:
          break

        # TODO(mbarbella): Optimize for empty script case (e.g. for "src=").
        token = HTMLMinimizer.Token(data[current_token_start:index],
                                    current_token_type)
        tokens.append(token)

        current_token_start = index
        current_token_type = HTMLMinimizer.Token.TYPE_HTML
        state = HTMLMinimizer.TokenizerState.SEARCHING_FOR_SCRIPT

    token = HTMLMinimizer.Token(data[current_token_start:], current_token_type)
    tokens.append(token)
    return tokens

  @staticmethod
  def combine_worker_tokens(tokens, prefix=b'', suffix=b''):
    """Combine tokens for a worker minimizer."""
    # The Antlr tokenizer decodes the bytes objects we originally pass to it.
    encoded_tokens = [
        t if isinstance(t, bytes) else t.encode('utf-8') for t in tokens
    ]
    return prefix + b''.join(encoded_tokens) + suffix

  @staticmethod
  def combine_tokens(tokens):
    """Combine tokens into a usable format, stripping metadata."""
    return b''.join([t.data for t in tokens])

  @staticmethod
  def run(data,
          thread_count=minimizer.DEFAULT_THREAD_COUNT,
          file_extension='.html'):
    """Attempt to minimize an html test case."""
    html_minimizer = HTMLMinimizer(
        utils.test, max_threads=thread_count, file_extension=file_extension)
    return html_minimizer.minimize(data)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for errors management."""

# Phrases to look in an error stacktrace to determine if something bad happened
# which is non recoverable and we just want to hang in that situation without
# proceeding further. This also helps to later come up and debug the problem.
BOT_ERROR_HANG_LIST = ['no space left']

# Phrases to look in an error stacktrace to determine whether the bot should
# terminate or not.
BOT_ERROR_TERMINATION_LIST = [
    'can\'t start new thread',
    'cannot allocate memory',
    'hostexception',
    'interrupted function call',
    'out of memory',
    'systemexit:',
]


def error_in_list(error_stacktrace, error_list):
  """Returns if the error exists in the error list."""
  # Change all strings to lowercase for comparison.
  error_stacktrace = error_stacktrace.lower()
  error_list = [error.lower() for error in error_list]

  for error in error_list:
    if error in error_stacktrace:
      return True

  return False


class Error(Exception):
  """Base exception class for errors."""


class InvalidTestcaseError(Error):
  """Error thrown when there is an attempt to access an invalid test case."""

  def __init__(self):
    super().__init__('Invalid test case.')


class InvalidFuzzerError(Error):
  """Error thrown when there is an attempt to set up a nonexistent fuzzer."""

  def __init__(self):
    super().__init__('Invalid fuzzer.')


class BadStateError(Error):
  """We are in an unexpected state that we cannot recover from."""

  def __init__(self, message=None):
    super().__init__('Entered a bad state.' or message)


class BuildNotFoundError(Error):
  """Exception type for build not found failures."""

  def __init__(self, revision, job_type):
    self.revision = revision
    self.job_type = job_type
    super().__init__()

  def __str__(self):
    return 'Build not found (revision %d, job %s).' % (self.revision,
                                                       self.job_type)


class BuildSetupError(Error):
  """Exception type for build setup failures."""

  def __init__(self, revision, job_type):
    self.revision = revision
    self.job_type = job_type
    super().__init__()

  def __str__(self):
    return 'Build setup failed (revision %d, job %s).' % (self.revision,
                                                          self.job_type)


class BadBuildError(Error):
  """Exception type for bad build failures."""

  def __init__(self, revision, job_type):
    self.revision = revision
    self.job_type = job_type
    super().__init__()

  def __str__(self):
    return 'Bad build detected (revision %d, job %s).' % (self.revision,
                                                          self.job_type)


class BadConfigError(Error):
  """Error thrown when configuration is bad."""

  def __init__(self, config_dir):
    super().__init__(
        'Bad configuration at: {config_dir}'.format(config_dir=config_dir))


class ConfigParseError(Error):
  """Error thrown when we failed to parse a config yaml file."""

  def __init__(self, file_path):
    self.file_path = file_path
    super().__init__()

  def __str__(self):
    return 'Failed to parse config file %s.' % self.file_path


class InvalidConfigKey(Error):
  """Error thrown when we failed to parse a config yaml file."""

  def __init__(self, key_name):
    self.key_name = key_name
    super().__init__()

  def __str__(self):
    return 'Invalid config key %s.' % self.key_name



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for the reproduce tool testcase_info handler."""
# pylint: disable=protected-access

import unittest

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.tests.test_libs import test_utils
from handlers.reproduce_tool import testcase_info


@test_utils.with_cloud_emulators('datastore')
class PrepareTestcaseDictTest(unittest.TestCase):
  """Tests for _prepare_testcase_dict."""

  def setUp(self):
    job = data_types.Job(name='test_job', environment_string='X = 1\nY = 2\n')
    job.put()

    testcase = data_types.Testcase()
    testcase.status = 'Pending'
    testcase.open = True
    testcase.job_type = 'test_job'
    testcase.put()

    self.testcase = testcase

  def test_expected_properties_included(self):
    """Ensure that a few of the common test case properties are included."""
    result = testcase_info._prepare_testcase_dict(self.testcase)
    self.assertEqual(result['status'], 'Pending')
    self.assertEqual(result['open'], True)
    self.assertEqual(result['group_id'], 0)

  def test_job_included(self):
    """Ensure that the job definition has been included."""
    result = testcase_info._prepare_testcase_dict(self.testcase)
    job_definition = result['job_definition']

    # Order is not necessarily preserved.
    self.assertIn('X = 1\n', job_definition)
    self.assertIn('Y = 2\n', job_definition)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Run minimizers with default configuration from the command line."""

import argparse
import os
import sys

from . import chunk_minimizer
from . import delta_minimizer
from . import html_minimizer
from . import js_minimizer
from . import minimizer
from . import utils


def main():
  """Minimize a file."""
  minimizers = {
      'chunk': chunk_minimizer.ChunkMinimizer,
      'html': html_minimizer.HTMLMinimizer,
      'js': js_minimizer.JSMinimizer,
      'line': delta_minimizer.DeltaMinimizer,
  }

  parser = argparse.ArgumentParser()
  parser.add_argument(
      '-t',
      '--threads',
      default=minimizer.DEFAULT_THREAD_COUNT,
      type=int,
      help='number of parallel instances')
  parser.add_argument(
      '-m',
      '--minimizer',
      choices=list(minimizers.keys()),
      default='line',
      help='minimization strategy')
  parser.add_argument(
      '-o', '--output-file', help='path to minimized output file')
  parser.add_argument(
      'COMMAND', help='command (quoted) to run for an individual test')
  parser.add_argument('FILE', help='file to minimize')
  args = vars(parser.parse_args(sys.argv[1:]))

  thread_count = args['threads']
  selected_minimizer = minimizers[args['minimizer']]
  command = args['COMMAND']
  file_path = args['FILE']
  file_extension = os.path.splitext(file_path)[1]
  output_file_path = args['output_file']
  if not output_file_path:
    output_file_path = '%s.min' % file_path

  utils.set_test_command(command)

  try:
    with open(file_path, 'rb') as file_handle:
      data = file_handle.read()
  except IOError:
    print('Unable to open input file %s.' % file_path)
    sys.exit(1)

  # Do not print an additional newline after minimization.
  minimized_output = selected_minimizer.run(
      data, thread_count=thread_count, file_extension=file_extension)
  print('Writing minimized output to %s.' % output_file_path)
  try:
    with open(output_file_path, 'wb') as file_handle:
      file_handle.write(minimized_output)
  except IOError:
    print('Unable to write output file %s.' % output_file_path)
    sys.exit(1)


if __name__ == '__main__':
  main()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler that gets the testcase list page."""

from flask import request

from clusterfuzz._internal.base import errors
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import blobs
from handlers import base_handler
from libs import crash_access
from libs import filters
from libs import handler
from libs import helpers
from libs.query import datastore_query

PAGE_SIZE = 20
MORE_LIMIT = 100 - PAGE_SIZE  # exactly 5 pages
FIELDS = [
    'crash_type',
    'crash_state',
    'job_type',
    'open',
    'fixed',
    'regression',
    'one_time_crasher_flag',
    'security_flag',
    'bug_information',
    'group_bug_information',
    'group_id',
    'project_name',
    'platform',
    'impact_extended_stable_version',
    'impact_stable_version',
    'impact_beta_version',
    'impact_head_version',
    'is_impact_set_flag',
]


class GroupFilter(filters.Filter):
  """Filter for group."""

  def __init__(self):
    self.param_key = 'group'

  def add(self, query, params):
    """Add group filter."""
    value = params.get(self.param_key, '')
    if filters.is_empty(value):
      query.filter('is_leader', True)
      return
    query.filter('group_id', helpers.cast(value, int, "'group' must be int."))


KEYWORD_FILTERS = [
    GroupFilter(),
    filters.String('bug_indices', 'issue'),
    filters.String('platform', 'platform'),
    filters.String('impact_extended_stable_version_indices', 'extended_stable'),
    filters.String('impact_head_version_indices', 'head'),
    filters.String('impact_stable_version_indices', 'stable'),
    filters.String('impact_beta_version_indices', 'beta'),
    filters.String('fuzzer_name_indices', 'fuzzer'),
    filters.String('job_type', 'job'),
]

FILTERS = [
    filters.String('impact_version_indices', 'impact'),
    filters.Boolean('has_bug_flag', 'issue'),
    filters.Boolean('open', 'open'),
    filters.Boolean('security_flag', 'security'),
    filters.Keyword(KEYWORD_FILTERS, 'keywords', 'q'),
    filters.NegativeBoolean('one_time_crasher_flag', 'reproducible'),
    filters.String('job_type', 'job'),
    filters.String('fuzzer_name_indices', 'fuzzer'),
    filters.String('project_name', 'project'),
    filters.Int('crash_revision', 'revision_greater_than', operator='>')
]


def add_filters(query, params):
  """Add filters based on params."""
  if not filters.has_params(params, FILTERS) and not params.get('showall'):
    params['open'] = 'yes'

  query.filter('status', 'Processed')
  query.filter('is_a_duplicate_flag', False)

  # For queries that use inequality we need to order by that field. Otherwise,
  # use the timestamp.
  if 'revision_greater_than' in params:
    query.order('crash_revision', is_desc=True)
  else:
    query.order('timestamp', is_desc=True)

  filters.add(query, params, FILTERS)


def get_result():
  """Get the result for the testcase list page."""
  params = dict(request.iterparams())
  page = helpers.cast(request.get('page') or 1, int, "'page' is not an int.")

  query = datastore_query.Query(data_types.Testcase)
  crash_access.add_scope(query, params, 'security_flag', 'job_type',
                         'fuzzer_name_indices')
  add_filters(query, params)

  testcases, total_pages, total_items, has_more = query.fetch_page(
      page=page, page_size=PAGE_SIZE, projection=FIELDS, more_limit=MORE_LIMIT)

  items = []
  for testcase in testcases:
    regression_range = ''
    fixed_range = ''

    if testcase.regression and testcase.regression != 'NA':
      regression_range = testcase.regression
    if testcase.fixed and testcase.fixed != 'NA':
      fixed_range = testcase.fixed

    item = {
        'id': testcase.key.id(),
        'crashType': ' '.join(testcase.crash_type.splitlines()),
        'crashStateLines': testcase.crash_state.strip().splitlines(),
        'jobType': testcase.job_type,
        'isClosed': not testcase.open,
        'isFixed': testcase.fixed and testcase.fixed != 'NA',
        'isReproducible': not testcase.one_time_crasher_flag,
        'isSecurity': testcase.security_flag,
        'isImpactSet': testcase.is_impact_set_flag,
        'impacts': {
            'extendedStable': testcase.impact_extended_stable_version,
            'stable': testcase.impact_stable_version,
            'beta': testcase.impact_beta_version,
            'head': testcase.impact_head_version,
        },
        'regressionRange': regression_range,
        'fixedRange': fixed_range,
        'groupId': testcase.group_id,
        'projectName': testcase.project_name,
        'platform': testcase.platform,
        'issueId': testcase.bug_information or testcase.group_bug_information,
        'showImpacts': testcase.has_impacts(),
        'impactsProduction': testcase.impacts_production()
    }
    if testcase.timestamp:
      item['timestamp'] = utils.utc_datetime_to_timestamp(testcase.timestamp)

    items.append(item)

  helpers.log('Testcases', helpers.VIEW_OPERATION)

  result = {
      'hasMore': has_more,
      'items': items,
      'page': page,
      'pageSize': PAGE_SIZE,
      'totalItems': total_items,
      'totalPages': total_pages,
  }
  return result, params


class Handler(base_handler.Handler):
  """Handler that gets the testcase list when user first lands on the page."""

  @handler.get(handler.HTML)
  def get(self):
    """Get and render the testcase list in HTML."""
    result, params = get_result()
    field_values = {
        'projects':
            data_handler.get_all_project_names(),
        'fuzzers':
            data_handler.get_all_fuzzer_names_including_children(
                include_parents=True),
        'jobs':
            data_handler.get_all_job_type_names(),
        'shouldShowImpact':
            utils.is_chromium()
    }
    return self.render('testcase-list.html', {
        'fieldValues': field_values,
        'result': result,
        'params': params
    })


class CacheHandler(base_handler.Handler):
  """Handler for exercising cache."""

  @handler.cron()
  def get(self):
    """Handle a GET request."""
    # pylint: disable=unexpected-keyword-arg

    # Memoize all project and job names.
    _ = data_handler.get_all_project_names(__memoize_force__=True)
    _ = data_handler.get_all_job_type_names(__memoize_force__=True)

    # Memoize both variants of get_all_fuzzer_names_including_children.
    _ = data_handler.get_all_fuzzer_names_including_children(
        include_parents=True, __memoize_force__=True)
    _ = data_handler.get_all_fuzzer_names_including_children(
        __memoize_force__=True)

    # Memoize expensive testcase attribute calls.
    for testcase_id in data_handler.get_open_testcase_id_iterator():
      try:
        testcase = data_handler.get_testcase_by_id(testcase_id)
      except errors.InvalidTestcaseError:
        # Already deleted.
        continue

      blobs.get_blob_size(testcase.fuzzed_keys)
      blobs.get_blob_size(testcase.minimized_keys)


class JsonHandler(base_handler.Handler):
  """Handler that gets the testcase list when user clicks on next page."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.oauth
  def post(self):
    """Get and render the testcase list in JSON."""
    result, _ = get_result()
    return self.render_json(result)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for remote_process."""

import os

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.bot.untrusted_runner import config
from clusterfuzz._internal.bot.untrusted_runner import file_impl
from clusterfuzz._internal.protos import untrusted_runner_pb2
from clusterfuzz._internal.tests.test_libs import test_utils


class FileImplTest(fake_filesystem_unittest.TestCase):
  """FileImpl tests."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)

  def test_create_directory(self):
    """Test file_impl.create_directory."""
    request = untrusted_runner_pb2.CreateDirectoryRequest(
        path='/dir', create_intermediates=False)
    response = file_impl.create_directory(request, None)
    self.assertTrue(response.result)
    self.assertTrue(os.path.isdir('/dir'))

    request = untrusted_runner_pb2.CreateDirectoryRequest(
        path='/dir2/dir2', create_intermediates=False)
    response = file_impl.create_directory(request, None)
    self.assertFalse(response.result)
    self.assertFalse(os.path.isdir('/dir2/dir2'))

    request = untrusted_runner_pb2.CreateDirectoryRequest(
        path='/dir3/dir3', create_intermediates=True)
    response = file_impl.create_directory(request, None)
    self.assertTrue(response.result)
    self.assertTrue(os.path.isdir('/dir3/dir3'))

  def test_remove_directory(self):
    """Test file_impl.remove_directory."""
    os.mkdir('/dir')
    request = untrusted_runner_pb2.RemoveDirectoryRequest(
        path='/dir', recreate=False)
    response = file_impl.remove_directory(request, None)
    self.assertTrue(response.result)
    self.assertFalse(os.path.isdir('/dir'))

    os.mkdir('/dir')
    request = untrusted_runner_pb2.RemoveDirectoryRequest(
        path='/dir', recreate=True)
    response = file_impl.remove_directory(request, None)
    self.assertTrue(response.result)
    self.assertTrue(os.path.isdir('/dir'))

  def test_copy_file_to_worker(self):
    """Test file_impl.copy_file_to_worker."""
    request_iterator = (
        untrusted_runner_pb2.FileChunk(data=b'A'),
        untrusted_runner_pb2.FileChunk(data=b'B'),
        untrusted_runner_pb2.FileChunk(data=b'C'),
    )

    context = mock.MagicMock()
    context.invocation_metadata.return_value = (('path-bin', b'/file'),)

    response = file_impl.copy_file_to_worker(request_iterator, context)
    self.assertTrue(response.result)
    self.assertTrue(os.path.exists('/file'))
    with open('/file') as f:
      self.assertEqual('ABC', f.read())

  def test_stat(self):
    """Test file_impl.stat."""
    self.fs.create_file('/file')
    request = untrusted_runner_pb2.StatRequest(path='/file')
    response = file_impl.stat(request, None)

    expected = os.stat('/file')
    self.assertTrue(response.result)
    self.assertEqual(expected.st_mode, response.st_mode)
    self.assertEqual(expected.st_size, response.st_size)
    self.assertEqual(expected.st_atime, response.st_atime)
    self.assertEqual(expected.st_ctime, response.st_ctime)
    self.assertEqual(expected.st_mtime, response.st_mtime)

  def test_stat_does_not_exist(self):
    """Test file_impl.stat (does not exist)."""
    request = untrusted_runner_pb2.StatRequest(path='/file')
    response = file_impl.stat(request, None)

    self.assertFalse(response.result)

  def test_copy_file_to_worker_create_intermediate(self):
    """Test file_impl.copy_file_to_worker (create intermediates)."""
    request_iterator = (
        untrusted_runner_pb2.FileChunk(data=b'A'),
        untrusted_runner_pb2.FileChunk(data=b'B'),
        untrusted_runner_pb2.FileChunk(data=b'C'),
    )

    context = mock.MagicMock()
    context.invocation_metadata.return_value = (('path-bin', b'/new_dir/file'),)

    response = file_impl.copy_file_to_worker(request_iterator, context)
    self.assertTrue(response.result)
    self.assertTrue(os.path.exists('/new_dir/file'))
    with open('/new_dir/file') as f:
      self.assertEqual('ABC', f.read())

  def test_copy_file_to_worker_create_dir_is_a_file(self):
    """Test file_impl.copy_file_to_worker when the directory is an existing
    file."""
    request_iterator = (
        untrusted_runner_pb2.FileChunk(data=b'A'),
        untrusted_runner_pb2.FileChunk(data=b'B'),
        untrusted_runner_pb2.FileChunk(data=b'C'),
    )

    self.fs.create_file('/file')

    context = mock.MagicMock()
    context.invocation_metadata.return_value = (('path-bin', b'/file/file'),)

    response = file_impl.copy_file_to_worker(request_iterator, context)
    self.assertFalse(response.result)
    self.assertTrue(os.path.isfile('/file'))

  def test_copy_file_to_worker_create_dir_error(self):
    """Test file_impl.copy_file_to_worker when we fail to create intermediate
    dirs."""
    request_iterator = (
        untrusted_runner_pb2.FileChunk(data=b'A'),
        untrusted_runner_pb2.FileChunk(data=b'B'),
        untrusted_runner_pb2.FileChunk(data=b'C'),
    )

    self.fs.create_file('/file')

    context = mock.MagicMock()
    context.invocation_metadata.return_value = (('path-bin',
                                                 b'/file/dir/file'),)

    response = file_impl.copy_file_to_worker(request_iterator, context)
    self.assertFalse(response.result)
    self.assertTrue(os.path.isfile('/file'))

  def test_copy_file_from_worker(self):
    """Test file_impl.copy_file_from_worker."""
    contents = (b'A' * config.FILE_TRANSFER_CHUNK_SIZE +
                b'B' * config.FILE_TRANSFER_CHUNK_SIZE +
                b'C' * config.FILE_TRANSFER_CHUNK_SIZE)

    self.fs.create_file('/file', contents=contents)

    request = untrusted_runner_pb2.CopyFileFromRequest(path='/file')
    context = mock.MagicMock()
    response = file_impl.copy_file_from_worker(request, context)

    chunks = [chunk.data for chunk in response]
    self.assertEqual(len(chunks), 3)
    self.assertEqual(contents, b''.join(chunks))
    context.set_trailing_metadata.assert_called_with([('result', 'ok')])

  def test_copy_file_from_worker_failed(self):
    """Test file_impl.copy_file_from_worker."""
    request = untrusted_runner_pb2.CopyFileFromRequest(path='/file')
    context = mock.MagicMock()
    response = file_impl.copy_file_from_worker(request, context)

    self.assertEqual(0, len(list(response)))
    context.set_trailing_metadata.assert_called_with([('result',
                                                       'invalid-path')])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""clean_indexes test."""
from collections import namedtuple
import os
import unittest

import mock

Args = namedtuple('Args', 'config_dir')

from clusterfuzz._internal.tests.test_libs import helpers
from local.butler import clean_indexes


class ExecuteTest(unittest.TestCase):
  """Execute tests."""

  def setUp(self):
    helpers.patch(self, ['local.butler.common.execute'])
    self.mock.execute.return_value = (0, 'ok')

  def test_invalid_config_dir(self):
    """Tests execute with invalid config dir."""
    args = Args(config_dir='')
    with self.assertRaises(SystemExit):
      clean_indexes.execute(args)

  def test_valid_config_dir(self):
    """Tests execute with valid config dir."""
    args = Args(config_dir=os.getenv('CONFIG_DIR_OVERRIDE'))
    clean_indexes.execute(args)

    self.mock.execute.assert_has_calls([
        mock.call('gcloud datastore indexes cleanup --quiet '
                  '--project test-clusterfuzz src/appengine/index.yaml'),
    ])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for performing remote tasks on windows."""
import os
import time

from fabric import api
from fabric import exceptions

from local.butler import common as butler_common
from local.butler import package
from local.remote import utils

TEMPLATE_REMMINA_PATH = 'src/local/remote/windows.remmina'
GENERATED_REMMINA_PATH = 'src/local/remote/generated.remmina'
EXTRACT_ZIP_PS_LOCAL_PATH = 'src/local/remote/extract_zip.ps1'
EXTRACT_ZIP_PS_REMOTE_PATH = r'c:\extract_clusterfuzz_stage_source_zip.ps1'


class Handler(object):
  """Handler for windows."""

  def __init__(self, instance_name, project, zone):
    assert project, 'Need to specify a project via --project argument.'
    assert zone, 'Need to specify a zone via --zone argument.'

    # FIXME: Make these configurable.
    self.username = 'clusterfuzz'
    self.domain = 'GOOGLE'
    self.clusterfuzz_parent_path = r'c:'

    self.staging_source_filename = 'clusterfuzz-source-stage.zip'

    (hostname, _, _) = utils.get_host_user_and_ssh_key_path(
        instance_name, project, zone)

    api.env.password = utils.get_password()
    api.env.user = self.username
    api.env.hosts = [hostname]
    api.env.host_string = '{username}@{hostname}'.format(
        username=self.username, hostname=hostname)
    api.env.key_filename = None
    api.env.gss_auth = None
    api.env.gss_deleg_creds = None
    api.env.gss_kex = None
    api.env.no_keys = True
    api.env.key_filename = None
    api.env.no_agent = True
    api.env.use_shell = False

  def _powershell(self, command, powershell_option='command'):
    """Wrap a command with powershell."""
    # Setting TMP is necessary to solve the flakiness problem.
    # See @tanin47's comment on:
    # https://github.com/PowerShell/PowerShell/issues/1746
    api.run(r'cmd /c set TMP=%USERPROFILE%\appdata\local\temp'
            ' && powershell -{powershell_option} {command}'.format(
                powershell_option=powershell_option, command=command))

  def _abspath(self, path):
    """Get absolute path on host given a path inside the clusterfuzz folder."""
    return self.clusterfuzz_parent_path + '\\clusterfuzz\\' + path

  def _log_path(self, log_name):
    return r'{log_dir}\{log_name}.log'.format(
        log_dir=self._abspath(r'bot\logs'), log_name=log_name)

  def tail(self, log_name, line_count):
    """Print the last `size` lines of ./bot/logs/`log_name`.log."""
    self._powershell(r'Get-Content -Path {log_path} -Tail {line_count}'.format(
        log_path=self._log_path(log_name), line_count=line_count))

  def tailf(self, log_names):
    """Print ./bot/logs/`name`.log in real-time (equivalent to `tail -f`)."""
    if len(log_names) > 1:
      raise Exception('Sorry, on windows, we cannot tailf multiple logs')

    self._powershell(r'Get-Content -Path {log_path} -Wait -Tail 100'.format(
        log_path=self._log_path(log_names[0])))

  def reboot(self):
    """Reboot the machine and verify if succeeded."""
    api.run('cmd /c shutdown /r /f /t 0')
    print('Sleeping 45 seconds to allow reboot to finish.')
    time.sleep(45)

    try:
      api.run('cmd /c echo "Test rebooting"')
      raise Exception(
          'Failed to reboot because we can still connect to the machine.')
    except exceptions.NetworkError:
      print('Cannot connect to the machine. The machine has been rebooted '
            'successfully.')

  def restart(self):
    """Kill all run_bot.py processes. run.py will get the new source code and
       start run_bot.py."""
    api.run('wmic process where'
            ' (name="python.exe" AND commandline like "%run_bot%")'
            ' get commandline,processid')
    api.run('wmic process where'
            ' (name="python.exe" AND commandline like "%run_bot%")'
            ' delete')

  def stage(self, config_dir):
    """Stage a zip (built by `python butler.py package`)."""
    os.environ['CONFIG_DIR_OVERRIDE'] = config_dir

    self.restart()
    time.sleep(1)

    zip_path = package.package(
        revision=butler_common.compute_staging_revision(),
        platform_name='windows')
    remote_zip_path = (
        '{clusterfuzz_parent_path}\\{staging_source_filename}'.format(
            clusterfuzz_parent_path=self.clusterfuzz_parent_path,
            staging_source_filename=self.staging_source_filename))
    api.put(zip_path, remote_zip_path)

    api.put(EXTRACT_ZIP_PS_LOCAL_PATH, EXTRACT_ZIP_PS_REMOTE_PATH)
    self._powershell(EXTRACT_ZIP_PS_REMOTE_PATH, 'file')

    self.restart()

  def rdp(self, share_path=None):
    """Launch remmina with correct configuration for the target instance."""
    password = utils.get_password()
    template = utils.get_file_content(TEMPLATE_REMMINA_PATH)

    with open(GENERATED_REMMINA_PATH, 'w') as config:
      config.write(
          template.format(
              domain=self.domain,
              username=self.username,
              password=password,
              share_path=share_path or '',
              hostname=api.env.hosts[0]))

    api.local('remmina -c ' + os.path.abspath(GENERATED_REMMINA_PATH))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for process."""

import os

import mock
from pyfakefs import fake_filesystem_unittest

from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import minijail
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers


class MinijailTest(fake_filesystem_unittest.TestCase):
  """Minijail tests."""

  def setUp(self):
    """Setup for minijail test."""
    if environment.platform() != 'LINUX':
      self.skipTest('Minijail tests are only applicable for linux platform.')

    self.setUpPyfakefs()
    for subdir in ['dev', 'lib', 'lib32', 'lib64', 'proc']:
      self.fs.create_dir(os.path.join('/', subdir))
    self.fs.create_dir(os.path.join('/', 'usr', 'lib'))
    self.fs.create_dir(os.path.join('/', 'usr', 'lib32'))

    test_helpers.patch(self, [
        'clusterfuzz._internal.system.minijail._get_minijail_path',
        'clusterfuzz._internal.system.minijail.MinijailChroot._mknod',
    ])

    test_helpers.patch_environ(self)

    self.mock._get_minijail_path.return_value = '/sbin/minijail'  # pylint: disable=protected-access

  def test_chroot(self):
    """Tests basic chroot setup."""
    chroot_directory = None
    with minijail.MinijailChroot() as chroot:
      chroot_directory = chroot.directory
      self.assertListEqual(
          sorted(os.listdir(chroot_directory)),
          ['dev', 'lib', 'lib32', 'lib64', 'proc', 'tmp', 'usr'])

      self.assertEqual(
          chroot.get_binding(chroot.tmp_directory),
          minijail.ChrootBinding(chroot.tmp_directory, '/tmp', True))

      for directory in ['/lib', '/lib32', '/lib64', '/usr/lib', '/usr/lib32']:
        self.assertEqual(
            chroot.get_binding(directory),
            minijail.ChrootBinding(directory, directory, False))

      self.assertIsNone(chroot.get_binding('/usr'))

    self.assertFalse(os.path.exists(chroot_directory))

  def test_chroot_bindings(self):
    """Tests chroot setup with additional bind dirs."""
    chroot_directory = None
    with minijail.MinijailChroot(bindings=[
        minijail.ChrootBinding('/foo/bar', '/bar', False),
    ]) as chroot:
      chroot_directory = chroot.directory
      self.assertListEqual(
          sorted(os.listdir(chroot_directory)),
          ['bar', 'dev', 'lib', 'lib32', 'lib64', 'proc', 'tmp', 'usr'])

    self.assertEqual(
        chroot.get_binding('/foo/bar'),
        minijail.ChrootBinding('/foo/bar', '/bar', False))
    self.assertFalse(os.path.exists(chroot_directory))

  @mock.patch('clusterfuzz._internal.system.minijail.os.getuid', lambda: 1000)
  def test_minijail(self):
    """Test minijail process command."""
    with minijail.MinijailChroot() as chroot:
      runner = minijail.MinijailProcessRunner(chroot, '/bin/ls')
      self.assertListEqual(runner.get_command(), [
          '/sbin/minijail', '-U', '-m', '0 1000 1', '-T', 'static', '-c', '0',
          '-n', '-v', '-p', '-l', '-I', '-k', 'proc,/proc,proc,1', '-P',
          chroot.directory, '-b',
          '%s,/tmp,1' % chroot.tmp_directory, '-b', '/lib,/lib,0', '-b',
          '/lib32,/lib32,0', '-b', '/lib64,/lib64,0', '-b',
          '/usr/lib,/usr/lib,0', '-b', '/usr/lib32,/usr/lib32,0', '/bin/ls'
      ])

  @mock.patch('clusterfuzz._internal.system.minijail.os.getuid', lambda: 1000)
  def test_minijail_bindings(self):
    """Test minijail process command with additional bind dirs."""
    with minijail.MinijailChroot(bindings=[
        minijail.ChrootBinding('/foo/bar', '/bar', True),
        minijail.ChrootBinding('/foo/barr', '/barr', False),
    ]) as chroot:
      runner = minijail.MinijailProcessRunner(chroot, '/bin/ls')
      self.assertListEqual(runner.get_command(), [
          '/sbin/minijail', '-U', '-m', '0 1000 1', '-T', 'static', '-c', '0',
          '-n', '-v', '-p', '-l', '-I', '-k', 'proc,/proc,proc,1', '-P',
          chroot.directory, '-b',
          '%s,/tmp,1' % chroot.tmp_directory, '-b', '/lib,/lib,0', '-b',
          '/lib32,/lib32,0', '-b', '/lib64,/lib64,0', '-b',
          '/usr/lib,/usr/lib,0', '-b', '/usr/lib32,/usr/lib32,0', '-b',
          '/foo/bar,/bar,1', '-b', '/foo/barr,/barr,0', '/bin/ls'
      ])

  @mock.patch('clusterfuzz._internal.system.minijail.os.getuid', lambda: 1000)
  @mock.patch('clusterfuzz._internal.system.minijail.subprocess.Popen')
  @mock.patch(
      'clusterfuzz._internal.system.minijail.tempfile.NamedTemporaryFile')
  def test_minijail_pid(self, mock_tempfile, _):
    """Test minijail process command writing to pid file."""
    mock_tempfile.return_value.name = '/temp_pid'

    with minijail.MinijailChroot() as chroot:
      runner = minijail.MinijailProcessRunner(chroot, 'bin/ls')
      process = runner.run()
      self.assertListEqual(process.command, [
          '/sbin/minijail', '-f', '/temp_pid', '-U', '-m', '0 1000 1', '-T',
          'static', '-c', '0', '-n', '-v', '-p', '-l', '-I', '-k',
          'proc,/proc,proc,1', '-P', chroot.directory, '-b',
          '%s,/tmp,1' % chroot.tmp_directory, '-b', '/lib,/lib,0', '-b',
          '/lib32,/lib32,0', '-b', '/lib64,/lib64,0', '-b',
          '/usr/lib,/usr/lib,0', '-b', '/usr/lib32,/usr/lib32,0', 'bin/ls'
      ])

  @mock.patch('clusterfuzz._internal.system.minijail.subprocess.Popen')
  def test_minijail_env_vars(self, mock_popen):
    """Test passing of env vars."""
    os.environ['ASAN_OPTIONS'] = 'asan_option=1'
    os.environ['AFL_OPTION'] = 'afl_option=1'
    os.environ['MSAN_OPTIONS'] = 'msan_option=1'
    os.environ['UBSAN_OPTIONS'] = 'ubsan_option=1'
    os.environ['SECRET'] = 'secret'
    os.environ['OTHER'] = 'other'

    with minijail.MinijailChroot() as chroot:
      runner = minijail.MinijailProcessRunner(chroot, 'binary')
      runner.run(env={'MSAN_OPTIONS': 'override=1', 'NAME': 'VALUE'})

      self.assertDictEqual({
          'MSAN_OPTIONS': 'override=1',
          'PATH': '/bin:/usr/bin',
      }, mock_popen.call_args[1]['env'])



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Fuzzing functions."""

from clusterfuzz._internal.bot.fuzzers import utils

from . import engine

_initialized = False

ENGINES = ('afl', 'googlefuzztest', 'hongfuzz', 'libFuzzer')


def _initialize():
  """Initialize the engine implementations."""
  global _initialized

  from clusterfuzz._internal.bot.fuzzers.afl import engine as afl_engine
  from clusterfuzz._internal.bot.fuzzers.googlefuzztest import \
      engine as gft_engine
  from clusterfuzz._internal.bot.fuzzers.honggfuzz import \
      engine as honggfuzz_engine
  from clusterfuzz._internal.bot.fuzzers.libFuzzer import \
      engine as libFuzzer_engine

  engine.register('afl', afl_engine.AFLEngine)
  engine.register('googlefuzztest', gft_engine.GoogleFuzzTestEngine)
  engine.register('honggfuzz', honggfuzz_engine.HonggfuzzEngine)
  engine.register('libFuzzer', libFuzzer_engine.LibFuzzerEngine)

  _initialized = True


def get_engine(name):
  """Get the engine with the given name."""
  if not _initialized:
    _initialize()

  engine_impl = engine.get(name)
  engine_impl.do_strategies = False
  return engine_impl


def is_fuzz_target(file_path, file_handle=None):
  """Returns whether |file_path| is a fuzz target."""
  return utils.is_fuzz_target_local(file_path, file_handle)


def get_fuzz_targets(directory):
  """Returns the list of fuzz targets in |directory|."""
  return utils.get_fuzz_targets_local(directory)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Format changed code in current branch."""

import os

from local.butler import common

FIRST_PARTY_MODULES = [
    'handlers',
    'libs',
    'clusterfuzz',
]

ISORT_CMD = ('isort --dont-order-by-type --force-single-line-imports '
             '--force-sort-within-sections --line-length=80 ' + ' '.join(
                 [f'-p {mod}' for mod in FIRST_PARTY_MODULES]) + ' ')


def execute(_):
  """Format changed code."""
  _, output = common.execute('git diff --name-only FETCH_HEAD')

  file_paths = [
      f.decode('utf-8') for f in output.splitlines() if os.path.exists(f)
  ]
  py_changed_file_paths = [
      f for f in file_paths if f.endswith('.py') and
      # Exclude auto-generated files.
      not f.endswith('_pb2.py') and not f.endswith('_pb2_grpc.py')
  ]
  go_changed_file_paths = [f for f in file_paths if f.endswith('.go')]
  for file_path in py_changed_file_paths:
    common.execute('yapf -i ' + file_path)
    common.execute(f'{ISORT_CMD} {file_path}')

  for file_path in go_changed_file_paths:
    common.execute('gofmt -w ' + file_path)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""big_query tests."""

import datetime
import unittest

import mock

from clusterfuzz._internal.datastore import data_types
from clusterfuzz._internal.google_cloud_utils import big_query
from clusterfuzz._internal.tests.test_libs import helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class InitTest(unittest.TestCase):
  """Test Client.__init__."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.get_application_id',
        'googleapiclient.discovery.build',
        'httplib2.Http',
    ])
    self.mock.get_application_id.return_value = 'project'
    self.mock.build.return_value = 'built'

  def test_init(self):
    """Test __init__."""
    client = big_query.Client(dataset_id='data', table_id='table')

    self.assertEqual('built', client.client)
    self.assertEqual('project', client.project_id)
    self.assertEqual('data', client.dataset_id)
    self.assertEqual('table', client.table_id)

    # `self.mock.build.assert_called_once_with(
    #      'bigquery', 'v2')` doesn't
    # work because `discovery.build` is decorated with @positional.
    # Therefore, we need the below.
    args, _ = self.mock.build.call_args
    self.assertEqual(('bigquery', 'v2'), args)


class RawQueryTest(unittest.TestCase):
  """Test Client.query."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.get_application_id',
        'clusterfuzz._internal.google_cloud_utils.big_query.get_api_client',
    ])
    self.mock.get_application_id.return_value = 'project'

  def test_query(self):
    """Test calling query API."""
    underlying = mock.MagicMock()
    jobs = mock.MagicMock()
    query = mock.MagicMock()

    underlying.jobs.return_value = jobs
    jobs.query.return_value = query
    query.execute.return_value = {'test': 1}
    self.mock.get_api_client.return_value = underlying

    client = big_query.Client()

    self.assertDictEqual({'test': 1}, client.raw_query('sql', max_results=100))
    jobs.query.assert_called_once_with(
        projectId='project',
        body={
            'query': 'sql',
            'timeoutMs': 60000,
            'maxResults': 100,
            'useLegacySql': False
        })


class InsertTest(unittest.TestCase):
  """Test Client.insert."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.get_application_id',
        'clusterfuzz._internal.google_cloud_utils.big_query.get_api_client',
    ])
    self.mock.get_application_id.return_value = 'project'

  def test_insert(self):
    """Test calling insertAll API."""
    underlying = mock.MagicMock()
    tabledata = mock.MagicMock()
    insert_all = mock.MagicMock()

    underlying.tabledata.return_value = tabledata
    tabledata.insertAll.return_value = insert_all
    insert_all.execute.return_value = {'test': 1}
    self.mock.get_api_client.return_value = underlying

    client = big_query.Client(dataset_id='data', table_id='table')

    self.assertDictEqual({
        'test': 1
    },
                         client.insert([
                             big_query.Insert({
                                 'a': 1
                             }, 'prefix:0'),
                             big_query.Insert({
                                 'b': 2
                             }, 'prefix:1')
                         ]))
    tabledata.insertAll.assert_called_once_with(
        projectId='project',
        datasetId='data',
        tableId='table',
        body={
            'kind':
                'bigquery#tableDataInsertAllRequest',
            'rows': [{
                'json': {
                    'a': 1
                },
                'insertId': 'prefix:0'
            }, {
                'json': {
                    'b': 2
                },
                'insertId': 'prefix:1'
            }]
        })


class ConvertTest(unittest.TestCase):
  """Test convert."""

  def test_convert(self):
    """Test convert every field."""
    fields = [{
        'name': 'int',
        'type': 'INTEGER',
        'mode': 'NULLABLE'
    }, {
        'name': 'float',
        'type': 'FLOAT',
        'mode': 'REQUIRED'
    }, {
        'name': 'bool',
        'type': 'BOOLEAN',
        'mode': 'NULLABLE'
    }, {
        'name': 'string',
        'type': 'STRING',
        'mode': 'REQUIRED'
    }, {
        'name': 'list-string',
        'type': 'STRING',
        'mode': 'REPEATED'
    }, {
        'name': 'time',
        'type': 'TIMESTAMP',
        'mode': 'REQUIRED'
    }, {
        'name':
            'counts',
        'type':
            'RECORD',
        'mode':
            'REPEATED',
        'fields': [{
            'name': 'name',
            'type': 'STRING',
            'mode': u'NULLABLE'
        }, {
            'fields': [{
                'name': 'hour',
                'type': 'INTEGER',
                'mode': 'NULLABLE'
            }, {
                'name': 'count',
                'type': 'INTEGER',
                'mode': u'NULLABLE'
            }],
            'type':
                'RECORD',
            'name':
                'slots',
            'mode':
                'REPEATED'
        }]
    }]
    result = {
        'schema': {
            'fields': fields
        },
        'rows': [{
            'f': [
                {
                    'v': '1'
                },
                {
                    'v': '2.3'
                },
                {
                    'v': 'true'
                },
                {
                    'v': 'str'
                },
                {
                    'v': [{
                        'v': 'a'
                    }, {
                        'v': 'b'
                    }]
                },
                {
                    'v': '1.485454187696014E9'
                },
                {
                    'v': [
                        {
                            'v': {
                                'f': [{
                                    'v': 'fuzzer1'
                                }, {
                                    'v': [
                                        {
                                            'v': {
                                                'f': [{
                                                    'v': 123
                                                }, {
                                                    'v': 999
                                                }]
                                            }
                                        },
                                        {
                                            'v': {
                                                'f': [{
                                                    'v': 124
                                                }, {
                                                    'v': 998
                                                }]
                                            }
                                        },
                                    ]
                                }]
                            }
                        },
                        {
                            'v': {
                                'f': [{
                                    'v': 'fuzzer2'
                                }, {
                                    'v': [
                                        {
                                            'v': {
                                                'f': [{
                                                    'v': 223
                                                }, {
                                                    'v': 899
                                                }]
                                            }
                                        },
                                        {
                                            'v': {
                                                'f': [{
                                                    'v': 224
                                                }, {
                                                    'v': 898
                                                }]
                                            }
                                        },
                                    ]
                                }]
                            }
                        },
                        {
                            'v': {
                                'f': [{
                                    'v': 'fuzzer3'
                                }, {
                                    'v': []
                                }]
                            }
                        },
                    ]
                },
            ],
        }, {
            'f': [{
                'v': '2'
            }, {
                'v': '3.3'
            }, {
                'v': 'false'
            }, {
                'v': 'str2'
            }, {
                'v': []
            }, {
                'v': '1.485455187696014E9'
            }, {
                'v': []
            }],
        }]
    }

    self.assertEqual([{
        'int':
            1,
        'float':
            2.3,
        'bool':
            True,
        'string':
            'str',
        'list-string': ['a', 'b'],
        'time':
            datetime.datetime.utcfromtimestamp(1485454187.696014),
        'counts': [
            {
                'name':
                    'fuzzer1',
                'slots': [
                    {
                        'count': 999,
                        'hour': 123
                    },
                    {
                        'count': 998,
                        'hour': 124
                    },
                ],
            },
            {
                'name':
                    'fuzzer2',
                'slots': [
                    {
                        'count': 899,
                        'hour': 223
                    },
                    {
                        'count': 898,
                        'hour': 224
                    },
                ],
            },
            {
                'name': 'fuzzer3',
                'slots': [],
            },
        ]
    }, {
        'int': 2,
        'float': 3.3,
        'bool': False,
        'string': 'str2',
        'list-string': [],
        'time': datetime.datetime.utcfromtimestamp(1485455187.696014),
        'counts': []
    }], big_query.convert(result))


class QueryTest(unittest.TestCase):
  """Test query."""

  def setUp(self):
    helpers.patch(self, [
        'clusterfuzz._internal.base.utils.get_application_id',
        'clusterfuzz._internal.google_cloud_utils.big_query.get_api_client',
        'time.time', 'time.sleep'
    ])
    self.mock.time.return_value = 1
    self.mock.get_application_id.return_value = 'project'

    underlying = mock.MagicMock()
    self.jobs = mock.MagicMock()

    self.query = mock.MagicMock()
    self.get_query_results = mock.MagicMock()

    underlying.jobs.return_value = self.jobs
    self.jobs.query.return_value = self.query
    self.jobs.getQueryResults.return_value = self.get_query_results

    self.mock.get_api_client.return_value = underlying

  def _make_resp(self, page_token, is_complete, total_row, count=1):
    """Make response."""
    resp = {
        'schema': {
            'fields': [{
                'name': 't',
                'type': 'INTEGER',
                'mode': 'NULLABLE'
            }]
        },
        'jobComplete': is_complete,
        'jobReference': {
            'jobId': 'job'
        },
        'totalRows': total_row
    }

    if is_complete:
      resp['rows'] = [{'f': [{'v': '1'}]}] * count

    if page_token:
      resp['pageToken'] = page_token

    return resp

  def test_one_page(self):
    """Test one page."""
    self.query.execute.return_value = self._make_resp(None, True, 1)
    self.get_query_results.execute.return_value = self._make_resp(None, True, 1)
    client = big_query.Client()

    result = client.query('sql', timeout=10, max_results=100)

    self.assertEqual([{'t': 1}], result.rows)
    self.assertEqual(1, result.total_count)
    self.query.execute.assert_called_once_with()
    self.get_query_results.execute.assert_called_once_with()
    self.jobs.query.assert_called_once_with(
        projectId='project',
        body={
            'query': 'sql',
            'timeoutMs': 60000,
            'useLegacySql': False,
            'maxResults': 0
        })

  def test_multiple_page(self):
    """Test multiple page."""
    self.query.execute.return_value = self._make_resp('tok', True, 2)
    self.get_query_results.execute.side_effect = [
        self._make_resp('tok2', True, 2),
        self._make_resp(None, True, 2)
    ]

    client = big_query.Client()

    result = client.query('sql', timeout=10, max_results=100, offset=50)

    self.assertEqual([{'t': 1}, {'t': 1}], result.rows)
    self.assertEqual(2, result.total_count)
    self.query.execute.assert_called_once_with()
    self.jobs.query.assert_called_once_with(
        projectId='project',
        body={
            'query': 'sql',
            'timeoutMs': 60000,
            'useLegacySql': False,
            'maxResults': 0
        })

    self.jobs.getQueryResults.assert_has_calls([
        mock.call(
            projectId='project',
            jobId='job',
            startIndex=50,
            pageToken=None,
            timeoutMs=60000,
            maxResults=100),
        mock.call().execute(),
        mock.call(
            projectId='project',
            jobId='job',
            startIndex=0,
            pageToken='tok2',
            timeoutMs=60000,
            maxResults=100),
        mock.call().execute()
    ])

  def test_multiple_page_with_limit(self):
    """Test multiple page with limit."""
    self.query.execute.return_value = self._make_resp('tok', True, 2)
    self.get_query_results.execute.side_effect = [
        self._make_resp('tok2', True, 2, 3),
        self._make_resp('tok3', True, 2, 3),
        self._make_resp('tok4', True, 2, 3),
        self._make_resp(None, True, 2, 1)
    ]

    client = big_query.Client()

    result = client.query(
        'sql', timeout=10, max_results=100, offset=50, limit=10)

    self.assertEqual([{'t': 1}] * 10, result.rows)
    self.assertEqual(2, result.total_count)
    self.query.execute.assert_called_once_with()
    self.jobs.query.assert_called_once_with(
        projectId='project',
        body={
            'query': 'sql',
            'timeoutMs': 60000,
            'useLegacySql': False,
            'maxResults': 0
        })

    self.jobs.getQueryResults.assert_has_calls([
        mock.call(
            projectId='project',
            jobId='job',
            startIndex=50,
            pageToken=None,
            timeoutMs=60000,
            maxResults=10),
        mock.call().execute(),
        mock.call(
            projectId='project',
            jobId='job',
            startIndex=0,
            pageToken='tok2',
            timeoutMs=60000,
            maxResults=7),
        mock.call().execute(),
        mock.call(
            projectId='project',
            jobId='job',
            startIndex=0,
            pageToken='tok3',
            timeoutMs=60000,
            maxResults=4),
        mock.call().execute(),
        mock.call(
            projectId='project',
            jobId='job',
            startIndex=0,
            pageToken='tok4',
            timeoutMs=60000,
            maxResults=1),
        mock.call().execute()
    ])

  def test_not_complete(self):
    """Test jobComplete=false for the first few request."""
    self.query.execute.return_value = self._make_resp(None, False, 4)
    self.get_query_results.execute.side_effect = [
        self._make_resp(None, False, 4),
        self._make_resp('tok', True, 8),
        self._make_resp(None, True, 8)
    ]

    client = big_query.Client()

    result = client.query('sql', timeout=10, max_results=100, offset=50)

    self.assertEqual([{'t': 1}, {'t': 1}], result.rows)
    self.assertEqual(8, result.total_count)
    self.query.execute.assert_called_once_with()
    self.jobs.query.assert_called_once_with(
        projectId='project',
        body={
            'query': 'sql',
            'timeoutMs': 60000,
            'useLegacySql': False,
            'maxResults': 0
        })

    self.jobs.getQueryResults.assert_has_calls([
        mock.call(
            projectId='project',
            jobId='job',
            timeoutMs=60000,
            maxResults=100,
            pageToken=None,
            startIndex=50),
        mock.call().execute(),
        mock.call(
            projectId='project',
            jobId='job',
            timeoutMs=60000,
            maxResults=100,
            pageToken=None,
            startIndex=50),
        mock.call().execute(),
        mock.call(
            projectId='project',
            jobId='job',
            pageToken='tok',
            timeoutMs=60000,
            maxResults=100,
            startIndex=0),
        mock.call().execute(),
    ])

  def test_timeout(self):
    """Test timeout."""
    self.query.execute.return_value = self._make_resp(None, False, 3)

    self.count = 0

    def get_query_results(**unused_kwargs):
      self.count += 1
      if self.count >= 3:
        self.mock.time.return_value = 100
      return self._make_resp('tok%d' % self.count, False, 3)

    self.get_query_results.execute.side_effect = get_query_results

    client = big_query.Client()

    with self.assertRaises(Exception) as cm:
      client.query('sql', timeout=10, max_results=100, offset=50)

    self.assertEqual("Timeout: the query doesn't finish within 10 seconds.",
                     str(cm.exception))

    self.query.execute.assert_called_once_with()
    self.jobs.query.assert_called_once_with(
        projectId='project',
        body={
            'query': 'sql',
            'timeoutMs': 60000,
            'useLegacySql': False,
            'maxResults': 0
        })

    self.jobs.getQueryResults.assert_has_calls([
        mock.call(
            projectId='project',
            jobId='job',
            timeoutMs=60000,
            maxResults=100,
            pageToken=None,
            startIndex=50),
        mock.call().execute(),
    ] * 3)


@test_utils.with_cloud_emulators('datastore')
class WriteRangeTest(unittest.TestCase):
  """Test write_range."""

  def setUp(self):
    self.client = mock.Mock(spec_set=big_query.Client)
    helpers.patch(self, [
        'clusterfuzz._internal.metrics.logs.log_error',
        'clusterfuzz._internal.google_cloud_utils.big_query.Client',
        'time.time',
    ])
    self.mock.time.return_value = 99
    self.mock.Client.return_value = self.client

    self.testcase = data_types.Testcase(
        crash_type='type',
        crash_state='state',
        security_flag=True,
        fuzzer_name='libfuzzer',
        overridden_fuzzer_name='libfuzzer_pdf',
        job_type='some_job')
    self.testcase.put()

  def test_write(self):
    """Tests write."""
    self.client.insert.return_value = {}
    big_query.write_range('regressions', self.testcase, 'regression', 456, 789)

    self.assertEqual(0, self.mock.log_error.call_count)
    self.client.insert.assert_called_once_with([
        big_query.Insert(
            row={
                'testcase_id': str(self.testcase.key.id()),
                'crash_type': 'type',
                'crash_state': 'state',
                'security_flag': True,
                'parent_fuzzer_name': 'libfuzzer',
                'fuzzer_name': 'libfuzzer_pdf',
                'job_type': 'some_job',
                'created_at': 99,
                'regression_range_start': 456,
                'regression_range_end': 789,
            },
            insert_id='%s:456:789' % self.testcase.key.id())
    ])

  def test_error(self):
    """Tests error."""
    self.client.insert.return_value = {'insertErrors': ['exception']}
    big_query.write_range('regressions', self.testcase, 'regression', 456, 789)

    self.mock.log_error.assert_called_once_with(
        ("Ignoring error writing the testcase's regression range (%s) to "
         'BigQuery.' % self.testcase.key.id()),
        exception=mock.ANY)
    self.client.insert.assert_called_once_with([
        big_query.Insert(
            row={
                'testcase_id': str(self.testcase.key.id()),
                'crash_type': 'type',
                'crash_state': 'state',
                'security_flag': True,
                'parent_fuzzer_name': 'libfuzzer',
                'fuzzer_name': 'libfuzzer_pdf',
                'job_type': 'some_job',
                'created_at': 99,
                'regression_range_start': 456,
                'regression_range_end': 789,
            },
            insert_id='%s:456:789' % self.testcase.key.id())
    ])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for managing a mac."""

import os

from fabric import api

from local.remote import utils
from local.remote.handlers import posix


class Handler(posix.Handler):
  """Handler for managing a mac host."""

  def __init__(self, instance_name, *_):
    super(Handler, self).__init__(
        instance_name=instance_name, platform='macos', project=None, zone=None)

    # FIXME: Make these configurable.
    self.username = 'chrome-bot'
    self.clusterfuzz_parent_path = '/b'

    api.env.host_string = self.instance_name
    api.env.user = self.username
    api.env.gateway = None
    api.env.use_ssh_config = True
    api.env.ssh_config_path = '~/.ssh/config'
    api.env.key_filename = None
    api.env.use_shell = True
    api.env.password = api.env.sudo_password = utils.get_password()

    # Disable passwordless authentication, so we don't have to touch
    # a security key multiple times because it tries to access keys that
    # we don't need.
    os.environ['SSH_AUTH_SOCK'] = ''

    print()
    print('SSHing into a mac machine only works if you performs the below in'
          ' order:')
    print('(1) Your `~/.ssh/config` is up-to-date. You can get it from'
          ' go/chrome-lab-ssh-config')
    print('(2) You are a member of clusterfuzz-build-access (see'
          ' go/clusterfuzz-build-access)')
    print('(3) ControlMaster is enabled. Put the below'
          ' lines in your `~/.ssh/config`:')
    print()
    print('Match host *')
    print('  ControlMaster auto')
    print('  ControlPath ~/.ssh/ctrl-%C')
    print('  ControlPersist 6h')
    print()
    print('(4) You ran `prodaccess --chromegolo_ssh` recently.')
    print('(5) You sshed into *one* of the mac machines in the last'
          ' 6 hours.')
    print()
    print('Now you can enjoy sshing into any mac machine.')
    print()
